#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
import difflib
import inspect
import logging
import re
import traceback
from datetime import datetime
from threading import RLock

from collections.abc import Mapping
from multiprocessing.pool import ThreadPool

from errbot import CommandError
from errbot.flow import FlowExecutor, FlowRoot
from .backends.base import Backend, Room, Identifier, Message
from .storage import StoreMixin
from .streaming import Tee
from .templating import tenv
from .utils import split_string_after

log = logging.getLogger(__name__)


# noinspection PyAbstractClass
class ErrBot(Backend, StoreMixin):
    """ ErrBot is the layer taking care of commands management and dispatching.
    """
    __errdoc__ = """ Commands related to the bot administration """
    MSG_ERROR_OCCURRED = 'Computer says nooo. See logs for details'
    MSG_UNKNOWN_COMMAND = 'Unknown command: "%(command)s". '
    startup_time = datetime.now()

    def __init__(self, bot_config):
        log.debug("ErrBot init.")
        super().__init__(bot_config)
        self.bot_config = bot_config
        self.prefix = bot_config.BOT_PREFIX
        if bot_config.BOT_ASYNC:
            self.thread_pool = ThreadPool(bot_config.BOT_ASYNC_POOLSIZE)
            log.debug('created a thread pool of size %d.', bot_config.BOT_ASYNC_POOLSIZE)
        self.commands = {}  # the dynamically populated list of commands available on the bot
        self.re_commands = {}  # the dynamically populated list of regex-based commands available on the bot
        self.command_filters = []  # the dynamically populated list of filters
        self.MSG_UNKNOWN_COMMAND = 'Unknown command: "%(command)s". ' \
                                   'Type "' + bot_config.BOT_PREFIX + 'help" for available commands.'
        if bot_config.BOT_ALT_PREFIX_CASEINSENSITIVE:
            self.bot_alt_prefixes = tuple(prefix.lower() for prefix in bot_config.BOT_ALT_PREFIXES)
        else:
            self.bot_alt_prefixes = bot_config.BOT_ALT_PREFIXES
        self.repo_manager = None
        self.plugin_manager = None
        self.storage_plugin = None
        self._plugin_errors_during_startup = None
        self.flow_executor = FlowExecutor(self)
        self._gbl = RLock()  # this protects internal structures of this class

    def attach_repo_manager(self, repo_manager):
        self.repo_manager = repo_manager

    def attach_plugin_manager(self, plugin_manager):
        self.plugin_manager = plugin_manager

    def attach_storage_plugin(self, storage_plugin):
        # the storage_plugin is needed by the plugins
        self.storage_plugin = storage_plugin

    def initialize_backend_storage(self):
        """
        Initialize storage for the backend to use.
        """
        log.debug("Initializing backend storage")
        assert self.plugin_manager is not None
        assert self.storage_plugin is not None
        self.open_storage(self.storage_plugin, f'{self.mode}_backend')

    @property
    def all_commands(self):
        """Return both commands and re_commands together."""
        with self._gbl:
            newd = dict(**self.commands)
            newd.update(self.re_commands)
        return newd

    def _dispatch_to_plugins(self, method, *args, **kwargs):
        """
        Dispatch the given method to all active plugins.

        Will catch and log any exceptions that occur.

        :param method: The name of the function to dispatch.
        :param *args: Passed to the callback function.
        :param **kwargs: Passed to the callback function.
        """
        for plugin in self.plugin_manager.get_all_active_plugins():
            plugin_name = plugin.name
            log.debug('Triggering %s on %s.', method, plugin_name)
            # noinspection PyBroadException
            try:
                getattr(plugin, method)(*args, **kwargs)
            except Exception:
                log.exception('%s on %s crashed.', method, plugin_name)

    def send(self, identifier, text, in_reply_to=None, groupchat_nick_reply=False):
        """ Sends a simple message to the specified user.

            :param identifier:
                an identifier from build_identifier or from an incoming message
            :param in_reply_to:
                the original message the bot is answering from
            :param text:
                the markdown text you want to send
            :param groupchat_nick_reply:
                authorized the prefixing with the nick form the user
        """
        # protect a little bit the backends here
        if not isinstance(identifier, Identifier):
            raise ValueError("identifier should be an Identifier")

        msg = self.build_message(text)
        msg.to = identifier
        msg.frm = in_reply_to.to if in_reply_to else self.bot_identifier
        msg.parent = in_reply_to

        nick_reply = self.bot_config.GROUPCHAT_NICK_PREFIXED
        if isinstance(identifier, Room) and in_reply_to and (nick_reply or groupchat_nick_reply):
            self.prefix_groupchat_reply(msg, in_reply_to.frm)

        self.split_and_send_message(msg)

    def send_templated(self, identifier, template_name, template_parameters, in_reply_to=None,
                       groupchat_nick_reply=False):
        """ Sends a simple message to the specified user using a template.

            :param template_parameters: the parameters for the template.
            :param template_name: the template name you want to use.
            :param identifier:
                an identifier from build_identifier or from an incoming message, a room etc.
            :param in_reply_to:
                the original message the bot is answering from
            :param groupchat_nick_reply:
                authorized the prefixing with the nick form the user
        """
        text = self.process_template(template_name, template_parameters)
        return self.send(identifier, text, in_reply_to, groupchat_nick_reply)

    def split_and_send_message(self, msg):
        for part in split_string_after(msg.body, self.bot_config.MESSAGE_SIZE_LIMIT):
            partial_message = msg.clone()
            partial_message.body = part
            partial_message.partial = True
            self.send_message(partial_message)

    def send_message(self, msg):
        """
        This needs to be overridden by the backends with a super() call.

        :param msg: the message to send.
        :return: None
        """
        for bot in self.plugin_manager.get_all_active_plugins():
            # noinspection PyBroadException
            try:
                bot.callback_botmessage(msg)
            except Exception:
                log.exception("Crash in a callback_botmessage handler")

    def send_card(self, card):
        """
        Sends a card, this can be overriden by the backends *without* a super() call.

        :param card: the card to send.
        :return: None
        """
        self.send_templated(card.to, 'card', {'card': card})

    def send_simple_reply(self, msg, text, private=False, threaded=False):
        """Send a simple response to a given incoming message

        :param private: if True will force a response in private.
        :param threaded: if True and if the backend supports it, sends the response in a threaded message.
        :param text: the markdown text of the message.
        :param msg: the message you are replying to.
        """
        reply = self.build_reply(msg, text, private=private, threaded=threaded)
        if isinstance(reply.to, Room) and self.bot_config.GROUPCHAT_NICK_PREFIXED:
            self.prefix_groupchat_reply(reply, msg.frm)
        self.split_and_send_message(reply)

    def process_message(self, msg):
        """Check if the given message is a command for the bot and act on it.
        It return True for triggering the callback_messages on the .callback_messages on the plugins.

        :param msg: the incoming message.
        """
        # Prepare to handle either private chats or group chats

        frm = msg.frm
        text = msg.body
        if not hasattr(msg.frm, 'person'):
            raise Exception(f'msg.frm not an Identifier as it misses the "person" property.'
                            f' Class of frm : {msg.frm.__class__}.')

        username = msg.frm.person
        user_cmd_history = self.cmd_history[username]

        if msg.delayed:
            log.debug('Message from history, ignore it.')
            return False

        if self.is_from_self(msg):
            log.debug("Ignoring message from self.")
            return False

        log.debug('*** frm = %s', frm)
        log.debug('*** username = %s', username)
        log.debug('*** text = %s', text)

        suppress_cmd_not_found = self.bot_config.SUPPRESS_CMD_NOT_FOUND

        prefixed = False  # Keeps track whether text was prefixed with a bot prefix
        only_check_re_command = False  # Becomes true if text is determed to not be a regular command
        tomatch = text.lower() if self.bot_config.BOT_ALT_PREFIX_CASEINSENSITIVE else text
        if len(self.bot_config.BOT_ALT_PREFIXES) > 0 and tomatch.startswith(self.bot_alt_prefixes):
            # Yay! We were called by one of our alternate prefixes. Now we just have to find out
            # which one... (And find the longest matching, in case you have 'err' and 'errbot' and
            # someone uses 'errbot', which also matches 'err' but would leave 'bot' to be taken as
            # part of the called command in that case)
            prefixed = True
            longest = 0
            for prefix in self.bot_alt_prefixes:
                length = len(prefix)
                if tomatch.startswith(prefix) and length > longest:
                    longest = length
            log.debug('Called with alternate prefix "%s"', text[:longest])
            text = text[longest:]

            # Now also remove the separator from the text
            for sep in self.bot_config.BOT_ALT_PREFIX_SEPARATORS:
                # While unlikely, one may have separators consisting of
                # more than one character
                length = len(sep)
                if text[:length] == sep:
                    text = text[length:]
        elif msg.is_direct and self.bot_config.BOT_PREFIX_OPTIONAL_ON_CHAT:
            log.debug('Assuming "%s" to be a command because BOT_PREFIX_OPTIONAL_ON_CHAT is True', text)
            # In order to keep noise down we surpress messages about the command
            # not being found, because it's possible a plugin will trigger on what
            # was said with trigger_message.
            suppress_cmd_not_found = True
        elif not text.startswith(self.bot_config.BOT_PREFIX):
            only_check_re_command = True
        if text.startswith(self.bot_config.BOT_PREFIX):
            text = text[len(self.bot_config.BOT_PREFIX):]
            prefixed = True

        text = text.strip()
        text_split = text.split(' ')
        cmd = None
        command = None
        args = ''
        if not only_check_re_command:
            i = len(text_split)
            while cmd is None:
                command = '_'.join(text_split[:i])

                with self._gbl:
                    if command in self.commands:
                        cmd = command
                        args = ' '.join(text_split[i:])
                    else:
                        i -= 1
                if i == 0:
                    break

            if command == self.bot_config.BOT_PREFIX:  # we did "!!" so recall the last command
                if len(user_cmd_history):
                    cmd, args = user_cmd_history[-1]
                else:
                    return False  # no command in history
            elif command.isdigit():  # we did "!#" so we recall the specified command
                index = int(command)
                if len(user_cmd_history) >= index:
                    cmd, args = user_cmd_history[-index]
                else:
                    return False  # no command in history

        # Try to match one of the regex commands if the regular commands produced no match
        matched_on_re_command = False
        if not cmd:
            with self._gbl:
                if prefixed or (msg.is_direct and self.bot_config.BOT_PREFIX_OPTIONAL_ON_CHAT):
                    commands = dict(self.re_commands)
                else:
                    commands = {k: self.re_commands[k] for k in self.re_commands
                                if not self.re_commands[k]._err_command_prefix_required}

            for name, func in commands.items():
                if func._err_command_matchall:
                    match = list(func._err_command_re_pattern.finditer(text))
                else:
                    match = func._err_command_re_pattern.search(text)
                if match:
                    log.debug('Matching "%s" against "%s" produced a match.', text,
                              func._err_command_re_pattern.pattern)
                    matched_on_re_command = True
                    self._process_command(msg, name, text, match)
                else:
                    log.debug('Matching "%s" against "%s" produced no match.',
                              text, func._err_command_re_pattern.pattern)
        if matched_on_re_command:
            return True

        if cmd:
            self._process_command(msg, cmd, args, match=None)
        elif not only_check_re_command:
            log.debug("Command not found")
            for cmd_filter in self.command_filters:
                if getattr(cmd_filter, 'catch_unprocessed', False):
                    try:
                        reply = cmd_filter(msg, cmd, args, False, emptycmd=True)
                        if reply:
                            self.send_simple_reply(msg, reply)
                        # continue processing the other unprocessed cmd filters.
                    except Exception:
                        log.exception("Exception in a command filter command.")
        return True

    def _process_command_filters(self, msg, cmd, args, dry_run=False):
        try:
            for cmd_filter in self.command_filters:
                msg, cmd, args = cmd_filter(msg, cmd, args, dry_run)
                if msg is None:
                    return None, None, None
            return msg, cmd, args
        except Exception:
            log.exception("Exception in a filter command, blocking the command in doubt")
            return None, None, None

    def _process_command(self, msg, cmd, args, match):
        """Process and execute a bot command"""

        # first it must go through the command filters
        msg, cmd, args = self._process_command_filters(msg, cmd, args, False)
        if msg is None:
            log.info('Command %s blocked or deferred.', cmd)
            return

        frm = msg.frm
        username = frm.person
        user_cmd_history = self.cmd_history[username]

        log.info(f'Processing command "{cmd}" with parameters "{args}" from {frm}')

        if (cmd, args) in user_cmd_history:
            user_cmd_history.remove((cmd, args))  # Avoids duplicate history items

        with self._gbl:
            f = self.re_commands[cmd] if match else self.commands[cmd]

        if f._err_command_admin_only and self.bot_config.BOT_ASYNC:
            # If it is an admin command, wait until the queue is completely depleted so
            # we don't have strange concurrency issues on load/unload/updates etc...
            self.thread_pool.close()
            self.thread_pool.join()
            self.thread_pool = ThreadPool(self.bot_config.BOT_ASYNC_POOLSIZE)

        if f._err_command_historize:
            user_cmd_history.append((cmd, args))  # add it to the history only if it is authorized to be so

        # Don't check for None here as None can be a valid argument to str.split.
        # '' was chosen as default argument because this isn't a valid argument to str.split()
        if not match and f._err_command_split_args_with != '':
            try:
                if hasattr(f._err_command_split_args_with, "parse_args"):
                    args = f._err_command_split_args_with.parse_args(args)
                elif callable(f._err_command_split_args_with):
                    args = f._err_command_split_args_with(args)
                else:
                    args = args.split(f._err_command_split_args_with)
            except Exception as e:
                self.send_simple_reply(msg, f"Sorry, I couldn't parse your arguments. {e}")
                return

        if self.bot_config.BOT_ASYNC:
            result = self.thread_pool.apply_async(
                self._execute_and_send,
                [],
                {'cmd': cmd, 'args': args, 'match': match, 'msg': msg, 'template_name': f._err_command_template}
            )
            if f._err_command_admin_only:
                # Again, if it is an admin command, wait until the queue is completely
                # depleted so we don't have strange concurrency issues.
                result.wait()
        else:
            self._execute_and_send(cmd=cmd, args=args, match=match, msg=msg,
                                   template_name=f._err_command_template)

    @staticmethod
    def process_template(template_name, template_parameters):
        # integrated templating
        # The template needs to be set and the answer from the user command needs to be a mapping
        # If not just convert the answer to string.
        if template_name and isinstance(template_parameters, Mapping):
            return tenv().get_template(template_name + '.md').render(**template_parameters)

        # Reply should be all text at this point (See https://github.com/errbotio/errbot/issues/96)
        return str(template_parameters)

    def _execute_and_send(self, cmd, args, match, msg, template_name=None):
        """Execute a bot command and send output back to the caller

        :param cmd: The command that was given to the bot (after being expanded)
        :param args: Arguments given along with cmd
        :param match: A re.MatchObject if command is coming from a regex-based command, else None
        :param msg: The message object
        :param template_name: The name of the jinja template which should be used to render
            the markdown output, if any

        """
        private = cmd in self.bot_config.DIVERT_TO_PRIVATE
        threaded = cmd in self.bot_config.DIVERT_TO_THREAD
        commands = self.re_commands if match else self.commands
        try:
            with self._gbl:
                method = commands[cmd]
            # first check if we need to reattach a flow context
            flow, _ = self.flow_executor.check_inflight_flow_triggered(cmd, msg.frm)
            if flow:
                log.debug("Reattach context from flow %s to the message", flow._root.name)
                msg.ctx = flow.ctx
            elif method._err_command_flow_only:
                # check if it is a flow_only command but we are not in a flow.
                log.debug("%s is tagged flow_only and we are not in a flow. Ignores the command.", cmd)
                return

            if inspect.isgeneratorfunction(method):
                replies = method(msg, match) if match else method(msg, args)
                for reply in replies:
                    if reply:
                        self.send_simple_reply(msg, self.process_template(template_name, reply), private, threaded)
            else:
                reply = method(msg, match) if match else method(msg, args)
                if reply:
                    self.send_simple_reply(msg, self.process_template(template_name, reply), private, threaded)

            # The command is a success, check if this has not made a flow progressed
            self.flow_executor.trigger(cmd, msg.frm, msg.ctx)

        except CommandError as command_error:
            reason = command_error.reason
            if command_error.template:
                reason = self.process_template(command_error.template, reason)
            self.send_simple_reply(msg, reason, private, threaded)

        except Exception as e:
            tb = traceback.format_exc()
            log.exception(f'An error happened while processing a message ("{msg.body}"): {tb}"')
            self.send_simple_reply(msg, self.MSG_ERROR_OCCURRED + f':\n{e}', private, threaded)

    def unknown_command(self, _, cmd, args):
        """ Override the default unknown command behavior
        """
        full_cmd = cmd + ' ' + args.split(' ')[0] if args else None
        if full_cmd:
            msg = f'Command "{cmd}" / "{full_cmd}" not found.'
        else:
            msg = f'Command "{cmd}" not found.'
        ununderscore_keys = [m.replace('_', ' ') for m in self.commands.keys()]
        matches = difflib.get_close_matches(cmd, ununderscore_keys)
        if full_cmd:
            matches.extend(difflib.get_close_matches(full_cmd, ununderscore_keys))
        matches = set(matches)
        if matches:
            alternatives = ('" or "' + self.bot_config.BOT_PREFIX).join(matches)
            msg += f'\n\nDid you mean "{self.bot_config.BOT_PREFIX}{alternatives}" ?'
        return msg

    def inject_commands_from(self, instance_to_inject):
        with self._gbl:
            plugin_name = instance_to_inject.name
            for name, value in inspect.getmembers(instance_to_inject, inspect.ismethod):
                if getattr(value, '_err_command', False):
                    commands = self.re_commands if getattr(value, '_err_re_command') else self.commands
                    name = getattr(value, '_err_command_name')

                    if name in commands:
                        f = commands[name]
                        new_name = (plugin_name + '-' + name).lower()
                        self.warn_admins(f'{plugin_name}.{name} clashes with {type(f.__self__).__name__}.{f.__name__} '
                                         f'so it has been renamed {new_name}')
                        name = new_name
                        value.__func__._err_command_name = new_name  # To keep track of the renaming.
                    commands[name] = value

                    if getattr(value, '_err_re_command'):
                        log.debug('Adding regex command : %s -> %s.', name, value.__name__)
                        self.re_commands = commands
                    else:
                        log.debug('Adding command : %s -> %s.', name, value.__name__)
                        self.commands = commands

    def inject_flows_from(self, instance_to_inject):
        classname = instance_to_inject.__class__.__name__
        for name, method in inspect.getmembers(instance_to_inject, inspect.ismethod):
            if getattr(method, '_err_flow', False):
                log.debug('Found new flow %s: %s', classname, name)
                flow = FlowRoot(name, method.__doc__)
                try:
                    method(flow)
                except Exception:
                    log.exception("Exception initializing a flow")

                self.flow_executor.add_flow(flow)

    def inject_command_filters_from(self, instance_to_inject):
        with self._gbl:
            for name, method in inspect.getmembers(instance_to_inject, inspect.ismethod):
                if getattr(method, '_err_command_filter', False):
                    log.debug('Adding command filter: %s', name)
                    self.command_filters.append(method)

    def remove_flows_from(self, instance_to_inject):
        for name, value in inspect.getmembers(instance_to_inject, inspect.ismethod):
            if getattr(value, '_err_flow', False):
                log.debug('Remove flow %s', name)
                # TODO(gbin)

    def remove_commands_from(self, instance_to_inject):
        with self._gbl:
            for name, value in inspect.getmembers(instance_to_inject, inspect.ismethod):
                if getattr(value, '_err_command', False):
                    name = getattr(value, '_err_command_name')
                    if getattr(value, '_err_re_command') and name in self.re_commands:
                        del self.re_commands[name]
                    elif not getattr(value, '_err_re_command') and name in self.commands:
                        del self.commands[name]

    def remove_command_filters_from(self, instance_to_inject):
        with self._gbl:
            for name, method in inspect.getmembers(instance_to_inject, inspect.ismethod):
                if getattr(method, '_err_command_filter', False):
                    log.debug('Removing command filter: %s', name)
                    self.command_filters.remove(method)

    def _admins_to_notify(self):
        """
        Creates a list of administrators to notify
        """
        admins_to_notify = self.bot_config.BOT_ADMINS_NOTIFICATIONS
        return admins_to_notify

    def warn_admins(self, warning: str) -> None:
        """
        Send a warning to the administrators of the bot.

        :param warning: The markdown-formatted text of the message to send.
        """
        for admin in self._admins_to_notify():
            self.send(self.build_identifier(admin), warning)
        log.warning(warning)

    def callback_message(self, msg):
        """Processes for commands and dispatches the message to all the plugins."""
        if self.process_message(msg):
            # Act only in the backend tells us that this message is OK to broadcast
            self._dispatch_to_plugins('callback_message', msg)

    def callback_mention(self, msg, people):
        log.debug("%s has/have been mentioned", ', '.join(str(p) for p in people))
        self._dispatch_to_plugins('callback_mention', msg, people)

    def callback_presence(self, pres):
        self._dispatch_to_plugins('callback_presence', pres)

    def callback_room_joined(self, room):
        """
            Triggered when the bot has joined a MUC.

            :param room:
                An instance of :class:`~errbot.backends.base.MUCRoom`
                representing the room that was joined.
        """
        self._dispatch_to_plugins('callback_room_joined', room)

    def callback_room_left(self, room):
        """
            Triggered when the bot has left a MUC.

            :param room:
                An instance of :class:`~errbot.backends.base.MUCRoom`
                representing the room that was left.
        """
        self._dispatch_to_plugins('callback_room_left', room)

    def callback_room_topic(self, room):
        """
            Triggered when the topic in a MUC changes.

            :param room:
                An instance of :class:`~errbot.backends.base.MUCRoom`
                representing the room for which the topic changed.
        """
        self._dispatch_to_plugins('callback_room_topic', room)

    def callback_stream(self, stream):
        log.info('Initiated an incoming transfer %s.', stream)
        Tee(stream, self.plugin_manager.get_all_active_plugins()).start()

    def signal_connect_to_all_plugins(self):
        for bot in self.plugin_manager.get_all_active_plugins():
            if hasattr(bot, 'callback_connect'):
                # noinspection PyBroadException
                try:
                    log.debug('Trigger callback_connect on %s.', bot.__class__.__name__)
                    bot.callback_connect()
                except Exception:
                    log.exception(f'callback_connect failed for {bot}.')

    def connect_callback(self):
        log.info('Activate internal commands')
        if self._plugin_errors_during_startup:
            errors = f'Some plugins failed to start during bot startup:\n\n{self._plugin_errors_during_startup}'
        else:
            errors = ''
        errors += self.plugin_manager.activate_non_started_plugins()
        if errors:
            self.warn_admins(errors)
            log.info(errors)
        log.info('Notifying connection to all the plugins...')
        self.signal_connect_to_all_plugins()
        log.info('Plugin activation done.')

    def disconnect_callback(self):
        log.info('Disconnect callback, deactivating all the plugins.')
        self.plugin_manager.deactivate_all_plugins()

    def get_doc(self, command):
        """Get command documentation
        """
        if not command.__doc__:
            return '(undocumented)'
        if self.prefix == '!':
            return command.__doc__
        ununderscore_keys = (m.replace('_', ' ') for m in self.all_commands.keys())
        pat = re.compile(fr'!({"|".join(ununderscore_keys)})')
        return re.sub(pat, self.prefix + '\1', command.__doc__)

    @staticmethod
    def get_plugin_class_from_method(meth):
        for cls in inspect.getmro(type(meth.__self__)):
            if meth.__name__ in cls.__dict__:
                return cls
        return None

    def get_command_classes(self):
        return (self.get_plugin_class_from_method(command)
                for command in self.all_commands.values())

    def shutdown(self):
        self.close_storage()
        self.plugin_manager.shutdown()
        self.repo_manager.shutdown()

    def prefix_groupchat_reply(self, message: Message, identifier: Identifier):
        if message.body.startswith('#'):
            # Markdown heading, insert an extra newline to ensure the
            # markdown rendering doesn't break.
            message.body = "\n" + message.body



import io
import json

from errbot import BotPlugin, botcmd, arg_botcmd
from errbot.flow import FlowNode, FlowRoot, Flow, FLOW_END
from errbot.core_plugins.acls import glob, get_acl_usr


class Flows(BotPlugin):
    """ Management commands related to flows / conversations.
    """

    def recurse_node(self, response: io.StringIO, stack, f: FlowNode, flow: Flow = None):
        if f in stack:
            response.write(f'{"&emsp;&nbsp;" * (len(stack))}↺<br>')
            return
        if isinstance(f, FlowRoot):
            doc = f.description if flow else ''
            response.write('Flow [' + f.name + '] ' + doc + ' <br>')
            if flow and flow.current_step == f:
                response.write(f'↪&nbsp;&nbsp;Start (_{flow.requestor}_)<br>')
        else:
            cmd = 'END' if f is FLOW_END else self._bot.all_commands[f.command]
            requestor = f'(_{str(flow.requestor)}_)' if flow and flow.current_step == f else ''
            doc = cmd.__doc__ if flow and f is not FLOW_END else ''
            response.write(f'{"&emsp;&nbsp;" * len(stack)}'
                           f'↪&nbsp;&nbsp;**{f if f is not FLOW_END else "END"}** {doc if doc else ""} {requestor}<br>')
        for _, sf in f.children:
            self.recurse_node(response, stack + [f], sf, flow)

    @botcmd(syntax='<name>')
    def flows_show(self, _, args):
        """ Shows the structure of a flow.
        """
        if not args:
            return 'You need to specify a flow name.'
        with io.StringIO() as response:
            flow_node = self._bot.flow_executor.flow_roots.get(args, None)
            if flow_node is None:
                return f"Flow {args} doesn't exist."
            self.recurse_node(response, [], flow_node)
            return response.getvalue()

    # noinspection PyUnusedLocal
    @botcmd
    def flows_list(self, msg, args):
        """ Displays the list of setup flows.
        """
        with io.StringIO() as response:
            for name, flow_node in self._bot.flow_executor.flow_roots.items():
                if flow_node.description:
                    response.write("- **" + name + "** " + flow_node.description + "\n")
                else:
                    response.write("- **" + name + "** " + "No description" + "\n")
            return response.getvalue()

    @botcmd(split_args_with=' ', syntax='<name> [initial_payload]')
    def flows_start(self, msg, args):
        """ Manually start a flow within the context of the calling user.
        You can prefeed the flow data with a json payload.
        Example:
             !flows start poll_setup {"title":"yeah!","options":["foo","bar","baz"]}
        """
        if not args:
            return 'You need to specify a flow to manually start'

        context = {}
        flow_name = args[0]
        if len(args) > 1:
            json_payload = ' '.join(args[1:])
            try:
                context = json.loads(json_payload)
            except Exception as e:
                return f'Cannot parse json {json_payload}: {e}.'
        self._bot.flow_executor.start_flow(flow_name, msg.frm, context)
        return f'Flow **{flow_name}** started ...'

    @botcmd()
    def flows_status(self, msg, args):
        """ Displays the list of started flows.
        """
        with io.StringIO() as response:
            if not self._bot.flow_executor.in_flight:
                response.write('No Flow started.\n')

            else:
                if not [flow for flow in self._bot.flow_executor.in_flight if self.check_user(msg, flow)]:
                    response.write(f'No Flow started for current user: {get_acl_usr(msg)}.\n')

                else:
                    if args:
                        for flow in self._bot.flow_executor.in_flight:
                            if self.check_user(msg, flow):
                                if flow.name == args:
                                    self.recurse_node(response, [], flow.root, flow)

                    else:
                        for flow in self._bot.flow_executor.in_flight:
                            if self.check_user(msg, flow):
                                next_steps = [f'\\*{str(step[1].command)}\\*' for step in
                                              flow._current_step.children if step[1].command]
                                next_steps_str = '\n'.join(next_steps)
                                text = f'\\>>> {str(flow.requestor)} is using flow \\*{flow.name}\\* on step ' \
                                       f'\\*{flow.current_step}\\*\nNext Step(s): \n{next_steps_str}'
                                response.write(text)
            return response.getvalue()

    @botcmd(syntax='[flow_name]')
    def flows_stop(self, msg, args):
        """ Stop flows you are in.
        optionally, stop a specific flow you are in.
        """
        if args:
            flow = self._bot.flow_executor.stop_flow(args, msg.frm)
            if flow:
                yield flow.name + ' stopped.'
                return
            yield 'Flow not found.'
            return

        one_stopped = False
        for flow in self._bot.flow_executor.in_flight:
            if flow.requestor == msg.frm:
                flow = self._bot.flow_executor.stop_flow(flow.name, msg.frm)
                if flow:
                    one_stopped = True
                    yield flow.name + ' stopped.'
        if not one_stopped:
            yield 'No Flow found.'

    @arg_botcmd('flow_name', type=str)
    @arg_botcmd('user', type=str)
    def flows_kill(self, _, user, flow_name):
        """ Admin command to kill a specific flow."""
        flow = self._bot.flow_executor.stop_flow(flow_name, self.build_identifier(user))
        if flow:
            return flow.name + ' killed.'
        return 'Flow not found.'

    def check_user(self, msg, flow):
        """Checks to make sure that either the user started the flow, or is a bot admin
        """
        if glob(get_acl_usr(msg), self.bot_config.BOT_ADMINS):
            return True
        elif glob(get_acl_usr(msg), flow.requestor.person):
            return True
        return False



import logging

import pytest

from errbot.backends.test import TestPerson
from errbot.flow import Flow, FlowRoot, InvalidState

log = logging.getLogger(__name__)


def test_node():
    root = FlowRoot("test", "This is my flowroot")
    node = root.connect("a", lambda ctx: ctx['toto'] == 'titui')

    assert root.predicate_for_node(node)({'toto': 'titui'})
    assert not root.predicate_for_node(node)({'toto': 'blah'})


def test_flow_predicate():
    root = FlowRoot("test", "This is my flowroot")
    node = root.connect("a", lambda ctx: 'toto' in ctx and ctx['toto'] == 'titui')
    somebody = TestPerson('me')

    # Non-matching predicate
    flow = Flow(root, somebody, {})
    assert node in flow.next_steps()
    assert node not in flow.next_autosteps()
    with pytest.raises(InvalidState):
        flow.advance(node)

    flow.advance(node, enforce_predicate=False)  # This will bypass the restriction
    assert flow._current_step == node

    # Matching predicate
    flow = Flow(root, somebody, {'toto': 'titui'})
    assert node in flow.next_steps()
    assert node in flow.next_autosteps()
    flow.advance(node)
    assert flow._current_step == node


def test_autotrigger():
    root = FlowRoot("test", "This is my flowroot")
    node = root.connect("a", lambda ctx: 'toto' in ctx and ctx['toto'] == 'titui', auto_trigger=True)
    assert node.command in root.auto_triggers



from ast import literal_eval
from pprint import pformat
import os
import shutil
import logging

from errbot import BotPlugin, botcmd
from errbot.plugin_manager import PluginConfigurationException, PluginActivationException
from errbot.repo_manager import RepoException


class Plugins(BotPlugin):

    @botcmd(admin_only=True)
    def repos_install(self, _, args):
        """ install a plugin repository from the given source or a known public repo (see !repos to find those).
        for example from a known repo : !install err-codebot
        for example a git url : git@github.com:gbin/plugin.git
        or an url towards a tar.gz archive : http://www.gootz.net/plugin-latest.tar.gz
        """
        args = args.strip()
        if not args:
            yield 'Please specify a repository listed in "!repos" or ' \
                  'give me the URL to a git repository that I should clone for you.'
            return
        try:
            yield f'Installing {args}...'
            local_path = self._bot.repo_manager.install_repo(args)
            errors = self._bot.plugin_manager.update_plugin_places(self._bot.repo_manager.get_all_repos_paths())
            if errors:
                v = '\n'.join(errors.values())
                yield f'Some plugins are generating errors:\n{v}.'
                # if the load of the plugin failed, uninstall cleanly teh repo
                for path in errors.keys():
                    if str(path).startswith(local_path):
                        yield f'Removing {local_path} as it did not load correctly.'
                        shutil.rmtree(local_path)
            else:
                yield f'A new plugin repository has been installed correctly from {args}. ' \
                      f'Refreshing the plugins commands...'
            loading_errors = self._bot.plugin_manager.activate_non_started_plugins()
            if loading_errors:
                yield loading_errors
            yield 'Plugins reloaded.'
        except RepoException as re:
            yield f'Error installing the repo: {re}'

    @botcmd(admin_only=True)
    def repos_uninstall(self, _, repo_name):
        """ uninstall a plugin repository by name.
        """
        if not repo_name.strip():
            yield "You should have a repo name as argument"
            return

        repos = self._bot.repo_manager.get_installed_plugin_repos()

        if repo_name not in repos:
            yield "This repo is not installed check with " + self._bot.prefix + "repos the list of installed ones"
            return

        plugin_path = os.path.join(self._bot.repo_manager.plugin_dir, repo_name)
        self._bot.plugin_manager.remove_plugins_from_path(plugin_path)
        self._bot.repo_manager.uninstall_repo(repo_name)
        yield f'Repo {repo_name} removed.'

    @botcmd(template='repos')
    def repos(self, _, args):
        """ list the current active plugin repositories
        """

        installed_repos = self._bot.repo_manager.get_installed_plugin_repos()

        all_names = [name for name in installed_repos]

        repos = {'repos': []}

        for repo_name in all_names:

            installed = False

            if repo_name in installed_repos:
                installed = True

            from_index = self._bot.repo_manager.get_repo_from_index(repo_name)

            if from_index is not None:
                description = '\n'.join((f'{plug.name}: {plug.documentation}' for plug in from_index))
            else:
                description = 'No description.'

            # installed, public, name, desc
            repos['repos'].append((installed, from_index is not None, repo_name, description))

        return repos

    @botcmd(template='repos2')
    def repos_search(self, _, args):
        """ Searches the repo index.
        for example: !repos search jenkins
        """
        if not args:
            # TODO(gbin): return all the repos.
            return {'error': "Please specify a keyword."}
        return {'repos': self._bot.repo_manager.search_repos(args)}

    @botcmd(split_args_with=' ', admin_only=True)
    def repos_update(self, _, args):
        """ update the bot and/or plugins
        use : !repos update all
        to update everything
        or : !repos update repo_name repo_name ...
        to update selectively some repos
        """
        if 'all' in args:
            results = self._bot.repo_manager.update_all_repos()
        else:
            results = self._bot.repo_manager.update_repos(args)

        yield "Start updating ... "

        for d, success, feedback in results:
            if success:
                yield f'Update of {d} succeeded...\n\n{feedback}\n\n'

                plugin = self._bot.plugin_manager.get_plugin_by_path(d)
                if hasattr(plugin, 'is_activated') and plugin.is_activated:
                    name = plugin.name
                    yield f'/me is reloading plugin {name}'
                    try:
                        self._bot.plugin_manager.reload_plugin_by_name(plugin.name)
                        yield f'Plugin {plugin.name} reloaded.'
                    except PluginActivationException as pae:
                        yield f'Error reactivating plugin {plugin.name}: {pae}'
            else:
                yield f'Update of {d} failed...\n\n{feedback}'

        yield "Done."

    @botcmd(split_args_with=' ', admin_only=True)
    def plugin_config(self, _, args):
        """ configure or get the configuration / configuration template for a specific plugin
        ie.
        !plugin config ExampleBot
        could return a template if it is not configured:
        {'LOGIN': 'example@example.com', 'PASSWORD': 'password', 'DIRECTORY': '/toto'}
        Copy paste, adapt so can configure the plugin :
        !plugin config ExampleBot {'LOGIN': 'my@email.com', 'PASSWORD': 'myrealpassword', 'DIRECTORY': '/tmp'}
        It will then reload the plugin with this config.
        You can at any moment retrieve the current values:
        !plugin config ExampleBot
        should return :
        {'LOGIN': 'my@email.com', 'PASSWORD': 'myrealpassword', 'DIRECTORY': '/tmp'}
        """
        plugin_name = args[0]
        if self._bot.plugin_manager.is_plugin_blacklisted(plugin_name):
            return f'Load this plugin first with {self._bot.prefix} load {plugin_name}.'
        obj = self._bot.plugin_manager.get_plugin_obj_by_name(plugin_name)
        if obj is None:
            return f'Unknown plugin or the plugin could not load {plugin_name}.'
        template_obj = obj.get_configuration_template()
        if template_obj is None:
            return 'This plugin is not configurable.'

        if len(args) == 1:
            response = f'Default configuration for this plugin (you can copy and paste this directly as a command):' \
                       f'\n\n```\n{self._bot.prefix}plugin config {plugin_name} {pformat(template_obj)}\n```'

            current_config = self._bot.plugin_manager.get_plugin_configuration(plugin_name)
            if current_config:
                response += f'\n\nCurrent configuration:\n\n```\n{self._bot.prefix}plugin config {plugin_name} ' \
                            f'{pformat(current_config)}\n```'
            return response

        # noinspection PyBroadException
        try:
            real_config_obj = literal_eval(' '.join(args[1:]))
        except Exception:
            self.log.exception('Invalid expression for the configuration of the plugin')
            return 'Syntax error in the given configuration'
        if type(real_config_obj) != type(template_obj):
            return 'It looks fishy, your config type is not the same as the template !'

        self._bot.plugin_manager.set_plugin_configuration(plugin_name, real_config_obj)

        try:
            self._bot.plugin_manager.deactivate_plugin(plugin_name)
        except PluginActivationException as pae:
            return f'Error deactivating {plugin_name}: {pae}.'

        try:
            self._bot.plugin_manager.activate_plugin(plugin_name)
        except PluginConfigurationException as ce:
            self.log.debug('Invalid configuration for the plugin, reverting the plugin to unconfigured.')
            self._bot.plugin_manager.set_plugin_configuration(plugin_name, None)
            return f'Incorrect plugin configuration: {ce}.'
        except PluginActivationException as pae:
            return f'Error activating plugin: {pae}.'

        return 'Plugin configuration done.'

    def formatted_plugin_list(self, active_only=True):
        """
        Return a formatted, plain-text list of loaded plugins.

        When active_only=True, this will only return plugins which
        are actually active. Otherwise, it will also include inactive
        (blacklisted) plugins.
        """
        if active_only:
            all_plugins = self._bot.plugin_manager.get_all_active_plugin_names()
        else:
            all_plugins = self._bot.plugin_manager.get_all_plugin_names()
        return "\n".join(("- " + plugin for plugin in all_plugins))

    @botcmd(admin_only=True)
    def plugin_reload(self, _, args):
        """reload a plugin: reload the code of the plugin leaving the activation status intact."""
        name = args.strip()
        if not name:
            yield (
                f'Please tell me which of the following plugins to reload:\n'
                f'{self.formatted_plugin_list(active_only=False)}')
            return
        if name not in self._bot.plugin_manager.get_all_plugin_names():
            yield (f'{name} isn\'t a valid plugin name. '
                   f'The current plugins are:\n{self.formatted_plugin_list(active_only=False)}')
            return

        if name not in self._bot.plugin_manager.get_all_active_plugin_names():
            answer = f'Warning: plugin {name} is currently not activated. '
            answer += f'Use `{self._bot.prefix}plugin activate {name}` to activate it.'
            yield answer

        try:
            self._bot.plugin_manager.reload_plugin_by_name(name)
            yield f'Plugin {name} reloaded.'
        except PluginActivationException as pae:
            yield f'Error activating plugin {name}: {pae}.'

    @botcmd(admin_only=True)
    def plugin_activate(self, _, args):
        """activate a plugin. [calls .activate() on the plugin]"""
        args = args.strip()
        if not args:
            return (f'Please tell me which of the following plugins to activate:\n'
                    f'{self.formatted_plugin_list(active_only=False)}')
        if args not in self._bot.plugin_manager.get_all_plugin_names():
            return (f"{args} isn't a valid plugin name. The current plugins are:\n"
                    f"{self.formatted_plugin_list(active_only=False)}")
        if args in self._bot.plugin_manager.get_all_active_plugin_names():
            return f'{args} is already activated.'

        try:
            self._bot.plugin_manager.activate_plugin(args)
        except PluginActivationException as pae:
            return f'Error activating plugin: {pae}'
        return f'Plugin {args} activated.'

    @botcmd(admin_only=True)
    def plugin_deactivate(self, _, args):
        """deactivate a plugin. [calls .deactivate on the plugin]"""
        args = args.strip()
        if not args:
            return (f'Please tell me which of the following plugins to deactivate:\n'
                    f'{self.formatted_plugin_list(active_only=False)}')
        if args not in self._bot.plugin_manager.get_all_plugin_names():
            return (f"{args} isn't a valid plugin name. The current plugins are:\n"
                    f"{self.formatted_plugin_list(active_only=False)}")
        if args not in self._bot.plugin_manager.get_all_active_plugin_names():
            return f'{args} is already deactivated.'

        try:
            self._bot.plugin_manager.deactivate_plugin(args)
        except PluginActivationException as pae:
            return f'Error deactivating {args}: {pae}'
        return f'Plugin {args} deactivated.'

    @botcmd(admin_only=True)
    def plugin_blacklist(self, _, args):
        """Blacklist a plugin so that it will not be loaded automatically during bot startup.
        If the plugin is currently activated, it will deactiveate it first."""
        if args not in self._bot.plugin_manager.get_all_plugin_names():
            return (f"{args} isn't a valid plugin name. The current plugins are:\n"
                    f"{self.formatted_plugin_list(active_only=False)}")

        if args in self._bot.plugin_manager.get_all_active_plugin_names():
            try:
                self._bot.plugin_manager.deactivate_plugin(args)
            except PluginActivationException as pae:
                return f'Error deactivating {args}: {pae}.'
        return self._bot.plugin_manager.blacklist_plugin(args)

    @botcmd(admin_only=True)
    def plugin_unblacklist(self, _, args):
        """Remove a plugin from the blacklist"""
        if args not in self._bot.plugin_manager.get_all_plugin_names():
            return (f"{args} isn't a valid plugin name. The current plugins are:\n"
                    f"{self.formatted_plugin_list(active_only=False)}")

        if args not in self._bot.plugin_manager.get_all_active_plugin_names():
            try:
                self._bot.plugin_manager.activate_plugin(args)
            except PluginActivationException as pae:
                return f'Error activating plugin: {pae}'

        return self._bot.plugin_manager.unblacklist_plugin(args)

    @botcmd(admin_only=True, template='plugin_info')
    def plugin_info(self, _, args):
        """Gives you a more technical information about a specific plugin."""
        pm = self._bot.plugin_manager
        if args not in pm.get_all_plugin_names():
            return (f"{args} isn't a valid plugin name. The current plugins are:\n"
                    f"{self.formatted_plugin_list(active_only=False)}")
        return {'plugin_info': pm.plugin_infos[args],
                'plugin': pm.plugins[args],
                'logging': logging}



import sys

import pytest
from io import StringIO
from pathlib import Path
from errbot.plugin_info import PluginInfo

plugfile_base = Path(__file__).absolute().parent / 'config_plugin'
plugfile_path = plugfile_base / 'config.plug'


def test_load_from_plugfile_path():
    pi = PluginInfo.load(plugfile_path)
    assert pi.name == 'Config'
    assert pi.module == 'config'
    assert pi.doc is None
    assert pi.python_version == (3, 0, 0)
    assert pi.errbot_minversion is None
    assert pi.errbot_maxversion is None


@pytest.mark.parametrize('test_input,expected', [
    ('2', (2, 0, 0)),
    ('2+', (3, 0, 0)),
    ('3', (3, 0, 0)),
    ('1.2.3', (1, 2, 3)),
    ('1.2.3-beta', (1, 2, 3)),
])
def test_python_version_parse(test_input, expected):
    f = StringIO("""
    [Core]
    Name = Config
    Module = config

    [Python]
    Version = %s
    """ % test_input)

    assert PluginInfo.load_file(f, None).python_version == expected


def test_doc():
    f = StringIO("""
    [Core]
    Name = Config
    Module = config

    [Documentation]
    Description = something
    """)

    assert PluginInfo.load_file(f, None).doc == 'something'


def test_errbot_version():
    f = StringIO("""
    [Core]
    Name = Config
    Module = config
    [Errbot]
    Min = 1.2.3
    Max = 4.5.6-beta
    """)
    info = PluginInfo.load_file(f, None)
    assert info.errbot_minversion == (1, 2, 3, sys.maxsize)
    assert info.errbot_maxversion == (4, 5, 6, 0)



from errbot import BotPlugin, cmdfilter


class CommandNotFoundFilter(BotPlugin):
    @cmdfilter(catch_unprocessed=True)
    def cnf_filter(self, msg, cmd, args, dry_run, emptycmd=False):
        """
        Check if command exists.  If not, signal plugins.  This plugin
        will be called twice: once as a command filter and then again
        as a "command not found" filter. See the emptycmd parameter.

        :param msg: Original chat message.
        :param cmd: Parsed command.
        :param args: Command arguments.
        :param dry_run: True when this is a dry-run.
        :param emptycmd: False when this command has been parsed and is valid.
        True if the command was not found.
        """

        if not emptycmd:
            return msg, cmd, args

        if self.bot_config.SUPPRESS_CMD_NOT_FOUND:
            self.log.debug('Suppressing command not found feedback.')
            return

        command = msg.body.strip()

        for prefix in self.bot_config.BOT_ALT_PREFIXES + (self.bot_config.BOT_PREFIX,):
            if command.startswith(prefix):
                command = command.replace(prefix, '', 1)
                break

        command_args = command.split(' ', 1)
        command = command_args[0]
        if len(command_args) > 1:
            args = ' '.join(command_args[1:])

        return self._bot.unknown_command(msg, command, args)



import os

extra_plugin_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'dependent_plugins')


def test_if_all_loaded_by_default(testbot):
    plug_names = testbot.bot.plugin_manager.get_all_active_plugin_names()
    assert 'Single' in plug_names
    assert 'Parent1' in plug_names
    assert 'Parent2' in plug_names


def test_single_dependency(testbot):
    pm = testbot.bot.plugin_manager
    for p in ('Single', 'Parent1', 'Parent2'):
        pm.deactivate_plugin(p)

    # everything should be gone
    plug_names = pm.get_all_active_plugin_names()
    assert 'Single' not in plug_names
    assert 'Parent1' not in plug_names
    assert 'Parent2' not in plug_names

    pm.activate_plugin('Single')

    # it should have activated the dependent plugin 'Parent1' only
    plug_names = pm.get_all_active_plugin_names()
    assert 'Single' in plug_names
    assert 'Parent1' in plug_names
    assert 'Parent2' not in plug_names


def test_double_dependency(testbot):
    pm = testbot.bot.plugin_manager
    all = ('Double', 'Parent1', 'Parent2')
    for p in all:
        pm.deactivate_plugin(p)

    pm.activate_plugin('Double')
    plug_names = pm.get_all_active_plugin_names()
    for p in all:
        assert p in plug_names


def test_dependency_retrieval(testbot):
    assert 'youpi' in testbot.exec_command('!depfunc')


def test_direct_circular_dependency(testbot):
    plug_names = testbot.bot.plugin_manager.get_all_active_plugin_names()
    assert 'Circular1' not in plug_names


def test_indirect_circular_dependency(testbot):
    plug_names = testbot.bot.plugin_manager.get_all_active_plugin_names()
    assert 'Circular2' not in plug_names
    assert 'Circular3' not in plug_names
    assert 'Circular4' not in plug_names



#!/usr/bin/env python

#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import sys

from setuptools import setup, find_packages

py_version = sys.version_info[:2]

if py_version < (3, 6):
    raise RuntimeError('Errbot requires Python 3.6 or later')

VERSION_FILE = os.path.join('errbot', 'version.py')

deps = ['webtest',
        'setuptools',
        'flask',
        'requests',
        'jinja2',
        'pyOpenSSL',
        'colorlog',
        'markdown<3.0',  # rendering stuff, 3.0+ deprecates 'safe()'
        'ansi',
        'Pygments>=2.0.2',
        'pygments-markdown-lexer>=0.1.0.dev39',  # sytax coloring to debug md
        'dulwich>=0.19.16'  # python implementation of git
        ]

src_root = os.curdir


def read_version():
    """
    Read directly the errbot/version.py and gives the version without loading Errbot.
    :return: errbot.version.VERSION
    """

    variables = {}
    with open(VERSION_FILE) as f:
        exec(compile(f.read(), 'version.py', 'exec'), variables)
    return variables['VERSION']


def read(fname, encoding='ascii'):
    return open(os.path.join(os.path.dirname(__file__), fname), 'r', encoding=encoding).read()


if __name__ == "__main__":

    VERSION = read_version()

    args = set(sys.argv)

    changes = read('CHANGES.rst', 'utf8')

    if changes.find(VERSION) == -1:
        raise Exception('You forgot to put a release note in CHANGES.rst ?!')

    if args & {'bdist', 'bdist_dumb', 'bdist_rpm', 'bdist_wininst', 'bdist_msi'}:
        raise Exception("err doesn't support binary distributions")

    packages = find_packages(src_root, include=['errbot', 'errbot.*'])

    setup(
        name="errbot",
        version=VERSION,
        packages=packages,
        entry_points={
            'console_scripts': [
                'errbot = errbot.cli:main',
            ]
        },

        install_requires=deps,
        tests_require=['nose', 'webtest', 'requests'],
        package_data={
            'errbot': ['backends/*.plug',
                       'backends/*.html',
                       'backends/styles/*.css',
                       'backends/images/*.svg',
                       'core_plugins/*.plug',
                       'core_plugins/*.md',
                       'core_plugins/templates/*.md',
                       'storage/*.plug',
                       'templates/initdir/example.py',
                       'templates/initdir/example.plug',
                       'templates/initdir/config.py.tmpl',
                       'templates/*.md',
                       'templates/new_plugin.py.tmpl',
                       ],
        },
        extras_require={
            'graphic': ['PySide', ],
            'hipchat': ['hypchat', 'sleekxmpp', 'pyasn1', 'pyasn1-modules'],
            'IRC': ['irc', ],
            'slack': ['slackclient>=1.0.5,<2.0', ],
            'slack-rtm': ['slackclient>=2.0', ],
            'telegram': ['python-telegram-bot', ],
            'XMPP': ['sleekxmpp', 'pyasn1', 'pyasn1-modules'],
            ':python_version<"3.7"': ['dataclasses'],  # backward compatibility for 3.3->3.6 for dataclasses
            ':sys_platform!="win32"': ['daemonize'],
        },

        author="errbot.io",
        author_email="info@errbot.io",
        description="Errbot is a chatbot designed to be simple to extend with plugins written in Python.",
        long_description=''.join([read('README.rst'), '\n\n', changes]),
        license="GPL",
        keywords="xmpp irc slack hipchat gitter tox chatbot bot plugin chatops",
        url="http://errbot.io/",
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Topic :: Communications :: Chat",
            "Topic :: Communications :: Chat :: Internet Relay Chat",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
        ],
        src_root=src_root,
        platforms='any',
    )



import sys
import threading
from json import JSONDecodeError
from urllib.error import HTTPError, URLError

import requests
from requests.exceptions import ConnectionError

from errbot import BotPlugin
from errbot.utils import version2tuple
from errbot.version import VERSION

HOME = 'https://errbot.io/versions.json'

installed_version = version2tuple(VERSION)

PY_VERSION = '.'.join(str(e) for e in sys.version_info[:3])


class VersionChecker(BotPlugin):

    connected = False
    activated = False

    def activate(self):
        if self.mode not in ('null', 'test', 'Dummy', 'text'):  # skip in all test confs.
            self.activated = True
            self.version_check()  # once at startup anyway
            self.start_poller(3600 * 24, self.version_check)  # once every 24H
            super().activate()
        else:
            self.log.info('Skip version checking under %s mode.', self.mode)

    def deactivate(self):
        self.activated = False
        super().deactivate()

    def _get_version(self):
        """Get errbot version based on python version."""
        version = VERSION
        major_py_version = PY_VERSION.partition('.')[0]

        # noinspection PyBroadException
        try:
            possible_versions = requests.get(HOME).json()
            version = possible_versions.get('python{}'.format(major_py_version), VERSION)
            self.log.debug("Latest Errbot version is: " + version)
        except (HTTPError, URLError, ConnectionError, JSONDecodeError):
            self.log.info('Could not establish connection to retrieve latest version.')
        return version

    def _async_vcheck(self):
        current_version_txt = self._get_version()
        self.log.debug("Installed Errbot version is: " + current_version_txt)
        current_version = version2tuple(current_version_txt)
        if installed_version < current_version:
            self.log.debug('A new version %s has been found, notify the admins!', current_version_txt)
            self.warn_admins(f'Version {current_version_txt} of Errbot is available. '
                             f'http://pypi.python.org/pypi/errbot/{current_version_txt}. '
                             f'To disable this check do: {self._bot.prefix}plugin blacklist VersionChecker')

    def version_check(self):
        if not self.activated:
            self.log.debug('Version check disabled')
            return
        self.log.debug('Checking version in background.')
        threading.Thread(target=self._async_vcheck).start()

    def callback_connect(self):
        if not self.connected:
            self.connected = True



from itertools import chain
from collections import namedtuple
from functools import partial
import io
import logging

from markdown import Markdown
from markdown.extensions import Extension
from markdown.postprocessors import Postprocessor
from markdown.inlinepatterns import SubstituteTagPattern
from markdown.extensions.fenced_code import FencedBlockPreprocessor

from ansi.colour import fg, bg, fx
from html import unescape


log = logging.getLogger(__name__)


# chr that should not count as a space
class NSC(object):
    def __init__(self, s):
        self.s = s

    def __str__(self):
        return self.s


# The translation table for the special characters.
CharacterTable = namedtuple('CharacterTable',
                            ['fg_black',
                             'fg_red',
                             'fg_green',
                             'fg_yellow',
                             'fg_blue',
                             'fg_magenta',
                             'fg_cyan',
                             'fg_white',
                             'fg_default',
                             'bg_black',
                             'bg_red',
                             'bg_green',
                             'bg_yellow',
                             'bg_blue',
                             'bg_magenta',
                             'bg_cyan',
                             'bg_white',
                             'bg_default',
                             'fx_reset',
                             'fx_bold',
                             'fx_italic',
                             'fx_underline',
                             'fx_not_italic',
                             'fx_not_underline',
                             'fx_normal',
                             'fixed_width',
                             'end_fixed_width',
                             'inline_code',
                             'end_inline_code',
                             ])

ANSI_CHRS = CharacterTable(fg_black=fg.black,
                           fg_red=fg.red,
                           fg_green=fg.green,
                           fg_yellow=fg.yellow,
                           fg_blue=fg.blue,
                           fg_magenta=fg.magenta,
                           fg_cyan=fg.cyan,
                           fg_white=fg.white,
                           fg_default=fg.default,
                           bg_black=bg.black,
                           bg_red=bg.red,
                           bg_green=bg.green,
                           bg_yellow=bg.yellow,
                           bg_blue=bg.blue,
                           bg_magenta=bg.magenta,
                           bg_cyan=bg.cyan,
                           bg_white=bg.white,
                           bg_default=bg.default,
                           fx_reset=fx.reset,
                           fx_bold=fx.bold,
                           fx_italic=fx.italic,
                           fx_underline=fx.underline,
                           fx_not_italic=fx.not_italic,
                           fx_not_underline=fx.not_underline,
                           fx_normal=fx.normal,
                           fixed_width='',
                           end_fixed_width='',
                           inline_code='',
                           end_inline_code='')


# Pure Text doesn't have any graphical chrs.
TEXT_CHRS = CharacterTable(fg_black='',
                           fg_red='',
                           fg_green='',
                           fg_yellow='',
                           fg_blue='',
                           fg_magenta='',
                           fg_cyan='',
                           fg_white='',
                           fg_default='',
                           bg_black='',
                           bg_red='',
                           bg_green='',
                           bg_yellow='',
                           bg_blue='',
                           bg_magenta='',
                           bg_cyan='',
                           bg_white='',
                           bg_default='',
                           fx_reset='',
                           fx_bold='',
                           fx_italic='',
                           fx_underline='',
                           fx_not_italic='',
                           fx_not_underline='',
                           fx_normal='',
                           fixed_width='',
                           end_fixed_width='',
                           inline_code='',
                           end_inline_code='')


# IMText have some formatting available
IMTEXT_CHRS = CharacterTable(fg_black='',
                             fg_red='',
                             fg_green='',
                             fg_yellow='',
                             fg_blue='',
                             fg_magenta='',
                             fg_cyan='',
                             fg_white='',
                             fg_default='',
                             bg_black='',
                             bg_red='',
                             bg_green='',
                             bg_yellow='',
                             bg_blue='',
                             bg_magenta='',
                             bg_cyan='',
                             bg_white='',
                             bg_default='',
                             fx_reset='',
                             fx_bold=NSC('*'),
                             fx_italic='',
                             fx_underline=NSC('_'),
                             fx_not_italic='',
                             fx_not_underline=NSC('_'),
                             fx_normal=NSC('*'),
                             fixed_width='```\n',
                             end_fixed_width='```\n',
                             inline_code='`',
                             end_inline_code='`')


NEXT_ROW = "&NEXT_ROW;"


class Table(object):

    def __init__(self, chr_table):
        self.headers = []
        self.rows = []
        self.in_headers = False
        self.ct = chr_table

    def next_row(self):
        if self.in_headers:
            self.headers.append([])  # is that exists ?
        else:
            self.rows.append([])

    def add_col(self):
        if not self.rows:
            self.rows = [[]]
        else:
            self.rows[-1].append(('', 0))

    def add_header(self):
        if not self.headers:
            self.headers = [[]]
        else:
            self.headers[-1].append(('', 0))

    def begin_headers(self):
        self.in_headers = True

    def end_headers(self):
        self.in_headers = False

    def write(self, text):
        cells = self.headers if self.in_headers else self.rows

        text_cell, count = cells[-1][-1]
        if isinstance(text, str):
            text_cell += text
            count += len(text)
        else:
            text_cell += str(text)  # This is a non space chr
        cells[-1][-1] = text_cell, count

    def __str__(self):
        nbcols = max(len(row) for row in chain(self.headers, self.rows))
        maxes = [0, ] * nbcols

        for row in chain(self.headers, self.rows):
            for i, el in enumerate(row):
                txt, length = el
                # Account for multiline cells
                cnt = str(txt).count(NEXT_ROW)
                if cnt > 0:
                    length -= cnt * len(NEXT_ROW)
                if maxes[i] < length:
                    maxes[i] = length

        # add up margins
        maxes = [m + 2 for m in maxes]

        output = io.StringIO()
        if self.headers:
            output.write('┏' + '┳'.join('━' * m for m in maxes) + '┓')
            output.write('\n')
            first = True
            for row in self.headers:
                if not first:
                    output.write('┣' + '╋'.join('━' * m for m in maxes) + '┫')
                    output.write('\n')
                first = False
                for i, header in enumerate(row):
                    text, ln = header
                    output.write('┃ ' + text + ' ' * (maxes[i] - 2 - ln) + ' ')
                output.write('┃')
                output.write('\n')
            output.write('┡' + '╇'.join('━' * m for m in maxes) + '┩')
            output.write('\n')
        else:
            output.write('┌' + '┬'.join('─' * m for m in maxes) + '┐')
            output.write('\n')
        first = True
        for row in self.rows:
            max_row_height = 1
            for i, item in enumerate(row):
                text, _ = item
                row_height = str(text).count(NEXT_ROW) + 1
                if row_height > max_row_height:
                    max_row_height = row_height
            if not first:
                output.write('├' + '┼'.join('─' * m for m in maxes) + '┤')
                output.write('\n')
            first = False
            for j in range(max_row_height):
                for i, item in enumerate(row):
                    text, ln = item
                    multi = text.split(NEXT_ROW)
                    if len(multi) > j:
                        text = multi[j]
                        ln = len(text)
                    else:
                        ln = 1
                        text = ' '
                    output.write('│ ' + text + ' ' * (maxes[i] - 2 - ln) + ' ')
                output.write('│')
                output.write('\n')
        output.write('└' + '┴'.join('─' * m for m in maxes) + '┘')
        output.write('\n')
        return str(self.ct.fixed_width) + output.getvalue() + str(self.ct.end_fixed_width)


class BorderlessTable(object):

    def __init__(self, chr_table):
        self.headers = []
        self.rows = []
        self.in_headers = False
        self.ct = chr_table

    def next_row(self):
        if self.in_headers:
            self.headers.append([])  # is that exists ?
        else:
            self.rows.append([])

    def add_col(self):
        if not self.rows:
            self.rows = [[]]
        else:
            self.rows[-1].append(('', 0))

    def add_header(self):
        if not self.headers:
            self.headers = [[]]
        else:
            self.headers[-1].append(('', 0))

    def begin_headers(self):
        self.in_headers = True

    def end_headers(self):
        self.in_headers = False

    def write(self, text):
        cells = self.headers if self.in_headers else self.rows

        text_cell, count = cells[-1][-1]
        if isinstance(text, str):
            text_cell += text
            count += len(text)
        else:
            text_cell += str(text)  # This is a non space chr
        cells[-1][-1] = text_cell, count

    def __str__(self):
        nbcols = max(len(row) for row in chain(self.headers, self.rows))
        maxes = [0, ] * nbcols

        for row in chain(self.headers, self.rows):
            for i, el in enumerate(row):
                txt, length = el
                # Account for multiline cells
                cnt = str(txt).count(NEXT_ROW)
                if cnt > 0:
                    length -= cnt * len(NEXT_ROW)
                if maxes[i] < length:
                    maxes[i] = length

        # add up margins
        maxes = [m + 2 for m in maxes]

        output = io.StringIO()
        if self.headers:
            for row in self.headers:
                for i, header in enumerate(row):
                    text, ln = header
                    output.write(text + ' ' * (maxes[i] - 2 - ln) + ' ')
                output.write('\n')
        for row in self.rows:
            max_row_height = 1
            for i, item in enumerate(row):
                text, _ = item
                row_height = str(text).count(NEXT_ROW) + 1
                if row_height > max_row_height:
                    max_row_height = row_height
            for j in range(max_row_height):
                for i, item in enumerate(row):
                    text, ln = item
                    multi = text.split(NEXT_ROW)
                    if len(multi) > j:
                        text = multi[j]
                        ln = len(text)
                    else:
                        ln = 1
                        text = ' '
                    output.write(text + ' ' * (maxes[i] - 2 - ln) + ' ')
                output.write('\n')
        return str(self.ct.fixed_width) + output.getvalue() + str(self.ct.end_fixed_width)


def recurse(write, chr_table, element, table=None, borders=True):
    post_element = []
    if element.text:
        text = element.text
    else:
        text = ''
    items = element.items()
    for k, v in items:
        if k == 'color':
            color_attr = getattr(chr_table, 'fg_' + v, None)
            if color_attr is None:
                log.warning("there is no '%s' color in ansi.", v)
                continue
            write(color_attr)
            post_element.append(chr_table.fg_default)
        elif k == 'bgcolor':
            color_attr = getattr(chr_table, 'bg_' + v, None)
            if color_attr is None:
                log.warning("there is no '%s' bgcolor in ansi", v)
                continue
            write(color_attr)
            post_element.append(chr_table.bg_default)
    if element.tag == 'img':
        text = dict(items)['src']
    elif element.tag == 'strong':
        write(chr_table.fx_bold)
        post_element.append(chr_table.fx_normal)
    elif element.tag == 'code':
        write(chr_table.inline_code)
        post_element.append(chr_table.end_inline_code)
    elif element.tag == 'em':
        write(chr_table.fx_underline)
        post_element.append(chr_table.fx_not_underline)
    elif element.tag == 'p':
        write(' ')
        post_element.append('\n')
    elif element.tag == 'br' and table:  # Treat <br/> differently in a table.
        write(NEXT_ROW)
    elif element.tag == 'a':
        post_element.append(' (' + element.get('href') + ')')
    elif element.tag == 'li':
        write('• ')
        post_element.append('\n')
    elif element.tag == 'hr':
        write('─' * 80)
        write('\n')
    elif element.tag == 'ul':  # ignore the text part
        text = None
    elif element.tag == 'h1':
        write(chr_table.fx_bold)
        text = text.upper()
        post_element.append(chr_table.fx_normal)
        post_element.append('\n\n')
    elif element.tag == 'h2':
        write('\n')
        write('  ')
        write(chr_table.fx_bold)
        post_element.append(chr_table.fx_normal)
        post_element.append('\n\n')
    elif element.tag == 'h3':
        write('\n')
        write('    ')
        write(chr_table.fx_underline)
        post_element.append(chr_table.fx_not_underline)
        post_element.append('\n\n')
    elif element.tag in ('h4', 'h5', 'h6'):
        write('\n')
        write('      ')
        post_element.append('\n')
    elif element.tag == 'table':
        table = Table(chr_table) if borders else BorderlessTable(chr_table)
        orig_write = write
        write = table.write
        text = None
    elif element.tag == 'tbody':
        text = None
    elif element.tag == 'thead':
        table.begin_headers()
        text = None
    elif element.tag == 'tr':
        table.next_row()
        text = None
    elif element.tag == 'td':
        table.add_col()
    elif element.tag == 'th':
        table.add_header()

    if text:
        write(text)
    for e in element:
        recurse(write, chr_table, e, table, borders)
    if element.tag == 'table':
        write = orig_write
        write(str(table))

    if element.tag == 'thead':
        table.end_headers()

    for restore in post_element:
        write(restore)
    if element.tail:
        tail = element.tail.rstrip('\n')
        if tail:
            write(tail)


def translate(element, chr_table=ANSI_CHRS, borders=True):
    f = io.StringIO()

    def write(ansi_obj):
        return f.write(str(ansi_obj))
    recurse(write, chr_table, element, borders=borders)
    result = f.getvalue().rstrip('\n')  # remove the useless final \n
    return result + str(chr_table.fx_reset)


# patch us in
def enable_format(name, chr_table, borders=True):
    Markdown.output_formats[name] = partial(translate, chr_table=chr_table, borders=borders)


for n, ct in (('ansi', ANSI_CHRS), ('text', TEXT_CHRS), ('imtext', IMTEXT_CHRS)):
    enable_format(n, ct)


class AnsiPostprocessor(Postprocessor):
    """Markdown generates html entities, this reputs them back to their unicode equivalent"""

    def run(self, text):
        return unescape(text)


# This is an adapted FencedBlockPreprocessor that doesn't insert <code><pre>
class AnsiPreprocessor(FencedBlockPreprocessor):
    def run(self, lines):
        """ Match and store Fenced Code Blocks in the HtmlStash. """
        text = "\n".join(lines)
        while 1:
            m = self.FENCED_BLOCK_RE.search(text)
            if m:
                code = self._escape(m.group('code'))

                placeholder = self.markdown.htmlStash.store(code)
                text = f'{text[:m.start()]}\n{placeholder}\n{text[m.end():]}'
            else:
                break
        return text.split('\n')

    def _escape(self, txt):
        """ basic html escaping """
        txt = txt.replace('&', '&amp;')
        txt = txt.replace('<', '&lt;')
        txt = txt.replace('>', '&gt;')
        txt = txt.replace('"', '&quot;')
        return txt


class AnsiExtension(Extension):
    """(kinda hackish) This is just a private extension to postprocess the html text to ansi text"""

    def extendMarkdown(self, md, md_globals):
        md.registerExtension(self)
        md.postprocessors.add(
            "unescape_html", AnsiPostprocessor(), ">unescape"
        )
        md.preprocessors.add(
            "ansi_fenced_codeblock", AnsiPreprocessor(md), "<fenced_code_block"
        )
        md.inlinePatterns.add(
            # Leave <br/> tags as is for proper table multiline cell processing
            "br", SubstituteTagPattern(r'<br/>', "br"), "<html"
        )
        del(md.preprocessors['fenced_code_block'])  # remove the old fenced block
        del(md.treeprocessors['prettify'])  # remove prettify treeprocessor since it adds extra new lines



import logging
from typing import Any
import shelve
import os

import shutil

from errbot.storage.base import StorageBase, StoragePluginBase

log = logging.getLogger('errbot.storage.shelf')


class ShelfStorage(StorageBase):
    def __init__(self, path):
        log.debug('Open shelf storage %s', path)
        self.shelf = shelve.DbfilenameShelf(path, protocol=2)

    def get(self, key: str) -> Any:
        return self.shelf[key]

    def remove(self, key: str):
        if key not in self.shelf:
            raise KeyError(f"{key} doesn't exist.")
        del self.shelf[key]

    def set(self, key: str, value: Any) -> None:
        self.shelf[key] = value

    def len(self):
        return len(self.shelf)

    def keys(self):
        return self.shelf.keys()

    def close(self) -> None:
        self.shelf.close()
        self.shelf = None


class ShelfStoragePlugin(StoragePluginBase):
    def __init__(self, bot_config):
        super().__init__(bot_config)
        if 'basedir' not in self._storage_config:
            self._storage_config['basedir'] = bot_config.BOT_DATA_DIR

    def open(self, namespace: str) -> StorageBase:
        config = self._storage_config
        # Hack to port move old DBs to the new location.
        new_spot = os.path.join(config['basedir'], namespace + '.db')
        old_spot = os.path.join(config['basedir'], 'plugins', namespace + '.db')
        if os.path.isfile(old_spot):
            if os.path.isfile(new_spot):
                log.warning('You have an old v3 DB at %s and a duplicate new one at %s.', old_spot, new_spot)
                log.warning('You need to either remove the old one or move it in place of the new one manually.')
            else:
                log.info('Moving your old v3 DB from %s to %s.', old_spot, new_spot)
                shutil.move(old_spot, new_spot)

        return ShelfStorage(new_spot)



import io
import logging
import random
import time
from typing import Any, Mapping, BinaryIO, List, Sequence, Tuple
from abc import ABC, abstractmethod
from collections import deque, defaultdict

log = logging.getLogger(__name__)


class Identifier(ABC):
    """This is just use for type hinting representing the Identifier contract,
    NEVER TRY TO SUBCLASS IT OUTSIDE OF A BACKEND, it is just here to show you what you can expect from an Identifier.
    To get an instance of a real identifier, always use the properties from Message (to, from) or self.build_identifier
     to make an identifier from a String.

     The semantics is anything you can talk to: Person, Room, RoomOccupant etc.
    """
    pass


class Person(Identifier):
    """This is just use for type hinting representing the Identifier contract,
    NEVER TRY TO SUBCLASS IT OUTSIDE OF A BACKEND, it is just here to show you what you can expect from an Identifier.
    To get an instance of a real identifier, always use the properties from Message (to, from) or self.build_identifier
     to make an identifier from a String.
    """

    @property
    @abstractmethod
    def person(self) -> str:
        """
        :return: a backend specific unique identifier representing the person you are talking to.
        """
        pass

    @property
    @abstractmethod
    def client(self) -> str:
        """
        :return: a backend specific unique identifier representing the device or client the person is using to talk.
        """
        pass

    @property
    @abstractmethod
    def nick(self) -> str:
        """
        :return: a backend specific nick returning the nickname of this person if available.
        """
        pass

    @property
    @abstractmethod
    def aclattr(self) -> str:
        """
        :return: returns the unique identifier that will be used for ACL matches.
        """
        pass

    @property
    @abstractmethod
    def fullname(self) -> str:
        """
        Some backends have the full name of a user.

        :return: the fullname of this user if available.
        """
        pass


class RoomOccupant(Identifier):
    @property
    @abstractmethod
    def room(self) -> Any:  # this is oom defined below
        """
        Some backends have the full name of a user.

        :return: the fullname of this user if available.
        """
        pass


class Room(Identifier):
    """
    This class represents a Multi-User Chatroom.
    """

    def join(self, username: str = None, password: str = None) -> None:
        """
        Join the room.

        If the room does not exist yet, this will automatically call
        :meth:`create` on it first.
        """
        raise NotImplementedError("It should be implemented specifically for your backend")

    def leave(self, reason: str = None) -> None:
        """
        Leave the room.

        :param reason:
            An optional string explaining the reason for leaving the room.
        """
        raise NotImplementedError("It should be implemented specifically for your backend")

    def create(self) -> None:
        """
        Create the room.

        Calling this on an already existing room is a no-op.
        """
        raise NotImplementedError("It should be implemented specifically for your backend")

    def destroy(self) -> None:
        """
        Destroy the room.

        Calling this on a non-existing room is a no-op.
        """
        raise NotImplementedError("It should be implemented specifically for your backend")

    @property
    def exists(self) -> bool:
        """
        Boolean indicating whether this room already exists or not.

        :getter:
            Returns `True` if the room exists, `False` otherwise.
        """
        raise NotImplementedError("It should be implemented specifically for your backend")

    @property
    def joined(self) -> bool:
        """
        Boolean indicating whether this room has already been joined.

        :getter:
            Returns `True` if the room has been joined, `False` otherwise.
        """
        raise NotImplementedError("It should be implemented specifically for your backend")

    @property
    def topic(self) -> str:
        """
        The room topic.

        :getter:
            Returns the topic (a string) if one is set, `None` if no
            topic has been set at all.

            .. note::
                Back-ends may return an empty string rather than `None`
                when no topic has been set as a network may not
                differentiate between no topic and an empty topic.
        :raises:
            :class:`~MUCNotJoinedError` if the room has not yet been joined.

        """
        raise NotImplementedError("It should be implemented specifically for your backend")

    @topic.setter
    def topic(self, topic: str) -> None:
        """
        Set the room's topic.

        :param topic:
            The topic to set.
        """
        raise NotImplementedError("It should be implemented specifically for your backend")

    @property
    def occupants(self) -> List[RoomOccupant]:
        """
        The room's occupants.

        :getter:
            Returns a list of occupant identities.
        :raises:
            :class:`~MUCNotJoinedError` if the room has not yet been joined.
        """
        raise NotImplementedError("It should be implemented specifically for your backend")

    def invite(self, *args) -> None:
        """
        Invite one or more people into the room.

        :*args:
            One or more identifiers to invite into the room.
        """
        raise NotImplementedError("It should be implemented specifically for your backend")


class RoomError(Exception):
    """General exception class for MUC-related errors"""


class RoomNotJoinedError(RoomError):
    """Exception raised when performing MUC operations
    that require the bot to have joined the room"""


class RoomDoesNotExistError(RoomError):
    """Exception that is raised when performing an operation
    on a room that doesn't exist"""


class UserDoesNotExistError(Exception):
    """Exception that is raised when performing an operation
    on a user that doesn't exist"""


class Message(object):
    """
    A chat message.

    This class represents chat messages that are sent or received by
    the bot.
    """

    def __init__(self,
                 body: str = '',
                 frm: Identifier = None,
                 to: Identifier = None,
                 parent: 'Message' = None,
                 delayed: bool = False,
                 partial: bool = False,
                 extras: Mapping = None,
                 flow=None):
        """
        :param body:
            The markdown body of the message.
        :param extras:
            Extra data attached by a backend
        :param flow:
            The flow in which this message has been triggered.
        :param parent:
            The parent message of this message in a thread. (Not supported by all backends)
        :param partial:
            Indicates whether the message was obtained by breaking down the message to fit
            the ``MESSAGE_SIZE_LIMIT``.
        """
        self._body = body
        self._from = frm
        self._to = to
        self._parent = parent
        self._delayed = delayed
        self._extras = extras or dict()
        self._flow = flow
        self._partial = partial

        # Convenience shortcut to the flow context
        if flow:
            self.ctx = flow.ctx
        else:
            self.ctx = {}

    def clone(self):
        return Message(body=self._body, frm=self._from, to=self._to, parent=self._parent,
                       delayed=self._delayed, partial=self._partial, extras=self._extras, flow=self._flow)

    @property
    def to(self) -> Identifier:
        """
        Get the recipient of the message.

        :returns:
            A backend specific identifier representing the recipient.
        """
        return self._to

    @to.setter
    def to(self, to: Identifier):
        """
        Set the recipient of the message.

        :param to:
            An identifier from for example build_identifier().
        """
        self._to = to

    @property
    def frm(self) -> Identifier:
        """
        Get the sender of the message.

        :returns:
            An :class:`~errbot.backends.base.Identifier` identifying
            the sender.
        """
        return self._from

    @frm.setter
    def frm(self, from_: Identifier):
        """
        Set the sender of the message.

        :param from_:
            An identifier from build_identifier.
        """
        self._from = from_

    @property
    def body(self) -> str:
        """
        Get the plaintext body of the message.

        :returns:
            The body as a string.
        """
        return self._body

    @body.setter
    def body(self, body: str):
        self._body = body

    @property
    def delayed(self) -> bool:
        return self._delayed

    @delayed.setter
    def delayed(self, delayed: bool):
        self._delayed = delayed

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent: 'Message'):
        self._parent = parent

    @property
    def extras(self) -> Mapping:
        return self._extras

    @property
    def flow(self):
        """
        Get the conversation flow for this message.

        :returns:
            A :class:`~errbot.Flow`
        """
        return self._from

    def __str__(self):
        return self._body

    @property
    def is_direct(self) -> bool:
        return isinstance(self.to, Person)

    @property
    def is_group(self) -> bool:
        return isinstance(self.to, Room)

    @property
    def is_threaded(self) -> bool:
        return self._parent is not None

    @property
    def partial(self) -> bool:
        return self._partial

    @partial.setter
    def partial(self, partial):
        self._partial = partial


class Card(Message):
    """
        Card is a special type of preformatted message. If it matches with a backend similar concept like on
        Slack or Hipchat it will be rendered natively, otherwise it will be sent as a regular message formatted with
        the card.md template.
    """

    def __init__(self,
                 body: str = '',
                 frm: Identifier = None,
                 to: Identifier = None,
                 parent: Message = None,
                 summary: str = None,
                 title: str = '',
                 link: str = None,
                 image: str = None,
                 thumbnail: str = None,
                 color: str = None,
                 fields: Tuple[Tuple[str, str]] = ()):
        """
        Creates a Card.
        :param body: main text of the card in markdown.
        :param frm: the card is sent from this identifier.
        :param to: the card is sent to this identifier (Room, RoomOccupant, Person...).
        :param parent: the parent message this card replies to. (threads the message if the backend supports it).
        :param summary: (optional) One liner summary of the card, possibly collapsed to it.
        :param title: (optional) Title possibly linking.
        :param link: (optional) url the title link is pointing to.
        :param image: (optional) link to the main image of the card.
        :param thumbnail: (optional) link to an icon / thumbnail.
        :param color: (optional) background color or color indicator.
        :param fields: (optional) a tuple of (key, value) pairs.
        """
        super().__init__(body=body, frm=frm, to=to, parent=parent)
        self._summary = summary
        self._title = title
        self._link = link
        self._image = image
        self._thumbnail = thumbnail
        self._color = color
        self._fields = fields

    @property
    def summary(self):
        return self._summary

    @property
    def title(self):
        return self._title

    @property
    def link(self):
        return self._link

    @property
    def image(self):
        return self._image

    @property
    def thumbnail(self):
        return self._thumbnail

    @property
    def color(self):
        return self._color

    @property
    def text_color(self):
        if self._color in ('black', 'blue'):
            return 'white'
        return 'black'

    @property
    def fields(self):
        return self._fields


ONLINE = 'online'
OFFLINE = 'offline'
AWAY = 'away'
DND = 'dnd'


class Presence(object):
    """
       This class represents a presence change for a user or a user in a chatroom.

       Instances of this class are passed to :meth:`~errbot.botplugin.BotPlugin.callback_presence`
       when the presence of people changes.
    """

    def __init__(self,
                 identifier: Identifier,
                 status: str = None,
                 message: str = None):
        if identifier is None:
            raise ValueError('Presence: identifiers is None')
        if status is None and message is None:
            raise ValueError('Presence: at least a new status or a new status message mustbe present')
        self._identifier = identifier
        self._status = status
        self._message = message

    @property
    def identifier(self) -> Identifier:
        """
        Identifier for whom its status changed. It can be a RoomOccupant or a Person.
        :return: the person or roomOccupant
        """
        return self._identifier

    @property
    def status(self) -> str:
        """ Returns the status of the presence change.
            It can be one of the constants ONLINE, OFFLINE, AWAY, DND, but
            can also be custom statuses depending on backends.
            It can be None if it is just an update of the status message (see get_message)
        """
        return self._status

    @property
    def message(self) -> str:
        """ Returns a human readable message associated with the status if any.
            like : "BRB, washing the dishes"
            It can be None if it is only a general status update (see get_status)
        """
        return self._message

    def __str__(self):
        response = ''
        if self._identifier:
            response += f'identifier: "{self._identifier}" '
        if self._status:
            response += f'status: "{self._status}" '
        if self._message:
            response += f'message: "{self._message}" '
        return response

    def __unicode__(self):
        return str(self.__str__())


STREAM_WAITING_TO_START = 'pending'
STREAM_TRANSFER_IN_PROGRESS = 'in progress'
STREAM_SUCCESSFULLY_TRANSFERED = 'success'
STREAM_PAUSED = 'paused'
STREAM_ERROR = 'error'
STREAM_REJECTED = 'rejected'

DEFAULT_REASON = 'unknown'


class Stream(io.BufferedReader):
    """
       This class represents a stream request.

       Instances of this class are passed to :meth:`~errbot.botplugin.BotPlugin.callback_stream`
       when an incoming stream is requested.
    """

    def __init__(self,
                 identifier: Identifier,
                 fsource: BinaryIO,
                 name: str = None,
                 size: int = None,
                 stream_type: str = None):
        super().__init__(fsource)
        self._identifier = identifier
        self._name = name
        self._size = size
        self._stream_type = stream_type
        self._status = STREAM_WAITING_TO_START
        self._reason = DEFAULT_REASON
        self._transfered = 0

    @property
    def identifier(self) -> Identifier:
        """
           The identity the stream is coming from if it is an incoming request
           or to if it is an outgoing request.
        """
        return self._identifier

    @property
    def name(self) -> str:
        """
            The name of the stream/file if it has one or None otherwise.
            !! Be carefull of injections if you are using this name directly as a filename.
        """
        return self._name

    @property
    def size(self) -> int:
        """
            The expected size in bytes of the stream if it is known or None.
        """
        return self._size

    @property
    def transfered(self) -> int:
        """
            The currently transfered size.
        """
        return self._transfered

    @property
    def stream_type(self) -> str:
        """
            The mimetype of the stream if it is known or None.
        """
        return self._stream_type

    @property
    def status(self) -> str:
        """
            The status for this stream.
        """
        return self._status

    def accept(self) -> None:
        """
            Signal that the stream has been accepted.
        """
        if self._status != STREAM_WAITING_TO_START:
            raise ValueError("Invalid state, the stream is not pending.")
        self._status = STREAM_TRANSFER_IN_PROGRESS

    def reject(self) -> None:
        """
            Signal that the stream has been rejected.
        """
        if self._status != STREAM_WAITING_TO_START:
            raise ValueError("Invalid state, the stream is not pending.")
        self._status = STREAM_REJECTED

    def error(self, reason=DEFAULT_REASON) -> None:
        """
            An internal plugin error prevented the transfer.
        """
        self._status = STREAM_ERROR
        self._reason = reason

    def success(self) -> None:
        """
            The streaming finished normally.
        """
        if self._status != STREAM_TRANSFER_IN_PROGRESS:
            raise ValueError("Invalid state, the stream is not in progress.")
        self._status = STREAM_SUCCESSFULLY_TRANSFERED

    def clone(self, new_fsource: BinaryIO) -> 'Stream':
        """
            Creates a clone and with an alternative stream
        """
        return Stream(self._identifier, new_fsource, self._name, self._size, self._stream_type)

    def ack_data(self, length: int) -> None:
        """ Acknowledge data has been transfered. """
        self._transfered = length


class Backend(ABC):
    """
    Implements the basic Bot logic (logic independent from the backend) and leaves
    you to implement the missing parts.
    """

    cmd_history = defaultdict(lambda: deque(maxlen=10))  # this will be a per user history

    MSG_ERROR_OCCURRED = 'Sorry for your inconvenience. ' \
                         'An unexpected error occurred.'

    def __init__(self, _):
        """ Those arguments will be directly those put in BOT_IDENTITY
        """
        log.debug("Backend init.")
        self._reconnection_count = 0  # Increments with each failed (re)connection
        self._reconnection_delay = 1  # Amount of seconds the bot will sleep on the
        #                                     # next reconnection attempt
        self._reconnection_max_delay = 600  # Maximum delay between reconnection attempts
        self._reconnection_multiplier = 1.75  # Delay multiplier
        self._reconnection_jitter = (0, 3)  # Random jitter added to delay (min, max)

    @abstractmethod
    def send_message(self, msg: Message) -> None:
        """Should be overridden by backends with a super().send_message() call."""

    @abstractmethod
    def change_presence(self, status: str = ONLINE, message: str = '') -> None:
        """Signal a presence change for the bot. Should be overridden by backends with a super().send_message() call."""

    @abstractmethod
    def build_reply(self, msg: Message, text: str = None, private: bool = False, threaded: bool = False):
        """ Should be implemented by the backend """

    @abstractmethod
    def callback_presence(self, presence: Presence) -> None:
        """ Implemented by errBot. """
        pass

    @abstractmethod
    def callback_room_joined(self, room: Room) -> None:
        """ See :class:`~errbot.errBot.ErrBot` """
        pass

    @abstractmethod
    def callback_room_left(self, room: Room) -> None:
        """ See :class:`~errbot.errBot.ErrBot` """
        pass

    @abstractmethod
    def callback_room_topic(self, room: Room) -> None:
        """ See :class:`~errbot.errBot.ErrBot` """
        pass

    def serve_forever(self) -> None:
        """
        Connect the back-end to the server and serve forever.

        Back-ends MAY choose to re-implement this method, in which case
        they are responsible for implementing reconnection logic themselves.

        Back-ends SHOULD trigger :func:`~connect_callback()` and
        :func:`~disconnect_callback()` themselves after connection/disconnection.
        """
        while True:
            try:
                if self.serve_once():
                    break  # Truth-y exit from serve_once means shutdown was requested
            except KeyboardInterrupt:
                log.info('Interrupt received, shutting down..')
                break
            except Exception:
                log.exception('Exception occurred in serve_once:')

            log.info('Reconnecting in %d seconds (%d attempted reconnections so far).', self._reconnection_delay,
                     self._reconnection_count)
            try:
                self._delay_reconnect()
                self._reconnection_count += 1
            except KeyboardInterrupt:
                log.info('Interrupt received, shutting down..')
                break

        log.info('Trigger shutdown')
        self.shutdown()

    def _delay_reconnect(self):
        """Delay next reconnection attempt until a suitable back-off time has passed"""
        time.sleep(self._reconnection_delay)

        self._reconnection_delay *= self._reconnection_multiplier
        if self._reconnection_delay > self._reconnection_max_delay:
            self._reconnection_delay = self._reconnection_max_delay
        self._reconnection_delay += random.uniform(*self._reconnection_jitter)  # nosec

    def reset_reconnection_count(self) -> None:
        """
        Reset the reconnection count. Back-ends should call this after
        successfully connecting.
        """
        self._reconnection_count = 0
        self._reconnection_delay = 1

    def build_message(self, text: str) -> Message:
        """ You might want to override this one depending on your backend """
        return Message(body=text)

    # ##### HERE ARE THE SPECIFICS TO IMPLEMENT PER BACKEND

    @abstractmethod
    def prefix_groupchat_reply(self, message: Message, identifier: Identifier):
        """ Patches message with the conventional prefix to ping the specific contact
        For example:
        @gbin, you forgot the milk !
        """

    @abstractmethod
    def build_identifier(self, text_representation: str) -> Identifier:
        pass

    def is_from_self(self, msg: Message) -> bool:
        """
        Needs to be overridden to check if the incoming message is from the bot itself.

        :param msg: The incoming message.
        :return: True if the message is coming from the bot.
        """
        # Default implementation (XMPP-like check using an extra config).
        # Most of the backends should have a better way to determine this.
        return (msg.is_direct and msg.frm == self.bot_identifier) or \
               (msg.is_group and msg.frm.nick == self.bot_config.CHATROOM_FN)

    def serve_once(self) -> None:
        """
        Connect the back-end to the server and serve a connection once
        (meaning until disconnected for any reason).

        Back-ends MAY choose not to implement this method, IF they implement a custom
        :func:`~serve_forever`.

        This function SHOULD raise an exception or return a value that evaluates
        to False in order to signal something went wrong. A return value that
        evaluates to True will signal the bot that serving is done and a shut-down
        is requested.
        """
        raise NotImplementedError("It should be implemented specifically for your backend")

    def connect(self) -> Any:
        """Connects the bot to server or returns current connection """

    @abstractmethod
    def query_room(self, room: str) -> Room:
        """
        Query a room for information.

        :param room:
            The room to query for.
        :returns:
            An instance of :class:`~Room`.
        """

    @abstractmethod
    def connect_callback(self) -> None:
        pass

    @abstractmethod
    def disconnect_callback(self) -> None:
        pass

    @property
    @abstractmethod
    def mode(self) -> str:
        pass

    @property
    @abstractmethod
    def rooms(self) -> Sequence[Room]:
        """
        Return a list of rooms the bot is currently in.

        :returns:
            A list of :class:`~errbot.backends.base.Room` instances.
        """



from typing import Any

from errbot.storage.base import StorageBase, StoragePluginBase

ROOTS = {}  # make a little bit of an emulated persistence.


class MemoryStorage(StorageBase):

    def __init__(self, namespace):
        self.namespace = namespace
        self.root = ROOTS.get(namespace, {})

    def get(self, key: str) -> Any:
        if key not in self.root:
            raise KeyError(f"{key} doesn't exist.")
        return self.root[key]

    def set(self, key: str, value: Any) -> None:
        self.root[key] = value

    def remove(self, key: str):
        if key not in self.root:
            raise KeyError(f"{key} doesn't exist.")
        del self.root[key]

    def len(self):
        return len(self.root)

    def keys(self):
        return self.root.keys()

    def close(self) -> None:
        ROOTS[self.namespace] = self.root


class MemoryStoragePlugin(StoragePluginBase):

    def open(self, namespace: str) -> StorageBase:
        return MemoryStorage(namespace)



# coding=utf-8
from datetime import timedelta
import pytest

from errbot.backends.test import ShallowConfig
from errbot.bootstrap import CORE_STORAGE, bot_config_defaults
from errbot.backend_plugin_manager import BackendPluginManager
from errbot.storage.base import StoragePluginBase
from errbot.utils import *
from errbot.storage import StoreMixin

log = logging.getLogger(__name__)


@pytest.mark.parametrize('v1,v2', [
                         ('2.0.0', '2.0.1'),
                         ('2.0.0', '2.1.0'),
                         ('2.0.0', '3.0.0'),
                         ('2.0.0-alpha', '2.0.0-beta'),
                         ('2.0.0-beta', '2.0.0-rc1'),
                         ('2.0.0-rc1', '2.0.0-rc2'),
                         ('2.0.0-rc2', '2.0.0-rc3'),
                         ('2.0.0-rc2', '2.0.0'),
                         ('2.0.0-beta', '2.0.1'),
                         ])
def test_version_check(v1, v2):
    assert version2tuple(v1) < version2tuple(v2)


@pytest.mark.parametrize('version', [
                         '1.2.3.4',
                         '1.2',
                         '1.2.-beta',
                         '1.2.3-toto',
                         '1.2.3-rc',
                         ])
def test_version_check_negative(version):
    with pytest.raises(ValueError):
        version2tuple(version)


def test_formattimedelta():
    td = timedelta(0, 60 * 60 + 13 * 60)
    assert '1 hours and 13 minutes' == format_timedelta(td)


def test_storage():
    key = 'test'

    __import__('errbot.config-template')
    config = ShallowConfig()
    config.__dict__.update(sys.modules['errbot.config-template'].__dict__)
    bot_config_defaults(config)

    spm = BackendPluginManager(config, 'errbot.storage', 'Memory', StoragePluginBase, CORE_STORAGE)
    storage_plugin = spm.load_plugin()

    persistent_object = StoreMixin()
    persistent_object.open_storage(storage_plugin, 'test')
    persistent_object[key] = 'à value'
    assert persistent_object[key] == 'à value'
    assert key in persistent_object
    del persistent_object[key]
    assert key not in persistent_object
    assert len(persistent_object) == 0


def test_split_string_after_returns_original_string_when_chunksize_equals_string_size():
    str_ = 'foobar2000' * 2
    splitter = split_string_after(str_, len(str_))
    split = [chunk for chunk in splitter]
    assert [str_] == split


def test_split_string_after_returns_original_string_when_chunksize_equals_string_size_plus_one():
    str_ = 'foobar2000' * 2
    splitter = split_string_after(str_, len(str_) + 1)
    split = [chunk for chunk in splitter]
    assert [str_] == split


def test_split_string_after_returns_two_chunks_when_chunksize_equals_string_size_minus_one():
    str_ = 'foobar2000' * 2
    splitter = split_string_after(str_, len(str_) - 1)
    split = [chunk for chunk in splitter]
    assert ['foobar2000foobar200', '0'] == split


def test_split_string_after_returns_two_chunks_when_chunksize_equals_half_length_of_string():
    str_ = 'foobar2000' * 2
    splitter = split_string_after(str_, int(len(str_) / 2))
    split = [chunk for chunk in splitter]
    assert ['foobar2000', 'foobar2000'] == split



from errbot import BotPlugin, botcmd

INROOM, USER, MULTILINE = 'inroom', 'user', 'multiline'


class TextModeCmds(BotPlugin):
    """
        Internal to TextBackend.
    """

    __errdoc__ = "Added commands for testing purposes"

    def activate(self):

        # This won't activate the plugin in anything else than text mode.
        if self.mode != 'text':
            return

        super().activate()

        # Some defaults if it was never used before'.
        if INROOM not in self:
            self[INROOM] = False

        if USER not in self:
            self[USER] = self.build_identifier(self.bot_config.BOT_ADMINS[0])

        if MULTILINE not in self:
            self[MULTILINE] = False

        # Restore the values to their live state.
        self._bot._inroom = self[INROOM]
        self._bot.user = self[USER]
        self._bot._multiline = self[MULTILINE]

    def deactivate(self):

        # Save the live state.
        self[INROOM] = self._bot._inroom
        self[USER] = self._bot.user
        self[MULTILINE] = self._bot._multiline

        super().deactivate()

    @botcmd
    def inroom(self, msg, args):
        """
           This puts you in a room with the bot.
        """
        self._bot._inroom = True
        if args:
            room = args
        else:
            room = '#testroom'
        self._bot.query_room(room).join()
        return f'Joined Room {room}.'

    @botcmd
    def inperson(self, msg, _):
        """
           This puts you in a 1-1 chat with the bot.
        """
        self._bot._inroom = False
        return 'Now in one-on-one with the bot.'

    @botcmd
    def asuser(self, msg, args):
        """
           This puts you in a room with the bot. You can specify a name otherwise it will default to 'luser'.
        """
        if args:
            usr = args
            if usr[0] != '@':
                usr = '@' + usr
            self._bot.user = self.build_identifier(usr)
        else:
            self._bot.user = self.build_identifier('@luser')
        return f'You are now: {self._bot.user}.'

    @botcmd
    def asadmin(self, msg, _):
        """
           This puts you in a 1-1 chat with the bot.
        """
        self._bot.user = self.build_identifier(self.bot_config.BOT_ADMINS[0])
        return f'You are now an admin: {self._bot.user}.'

    @botcmd
    def ml(self, msg, _):
        """
           Switch back and forth between normal mode and multiline mode. Use this if you want to test
           commands spanning multiple lines. Note: in multiline, press enter twice to end and send the message.
        """
        self._bot._multiline = not self._bot._multiline
        return 'Multiline mode, press enter twice to end messages' if self._bot._multiline else 'Normal one line mode.'



from os import path
from queue import Empty

import pytest

extra_plugin_dir = path.join(path.dirname(path.realpath(__file__)), 'flow_plugin')


def test_list_flows(testbot):
    assert len(testbot.bot.flow_executor.flow_roots) == 4
    testbot.bot.push_message('!flows list')
    result = testbot.pop_message()
    assert 'documentation of W1' in result
    assert 'documentation of W2' in result
    assert 'documentation of W3' in result
    assert 'documentation of W4' in result
    assert 'w1' in result
    assert 'w2' in result
    assert 'w3' in result
    assert 'w4' in result


def test_no_autotrigger(testbot):
    assert 'a' in testbot.exec_command('!a')
    assert len(testbot.bot.flow_executor.in_flight) == 0


def test_autotrigger(testbot):
    assert 'c' in testbot.exec_command('!c')
    flow_message = testbot.pop_message()
    assert 'You are in the flow w2, you can continue with' in flow_message
    assert '!b' in flow_message
    assert len(testbot.bot.flow_executor.in_flight) == 1
    assert testbot.bot.flow_executor.in_flight[0].name == 'w2'


def test_no_duplicate_autotrigger(testbot):
    assert 'c' in testbot.exec_command('!c')
    flow_message = testbot.pop_message()
    assert 'You are in the flow w2, you can continue with' in flow_message
    assert 'c' in testbot.exec_command('!c')
    assert len(testbot.bot.flow_executor.in_flight) == 1
    assert testbot.bot.flow_executor.in_flight[0].name == 'w2'


def test_secondary_autotrigger(testbot):
    assert 'e' in testbot.exec_command('!e')
    second_message = testbot.pop_message()
    assert 'You are in the flow w2, you can continue with' in second_message
    assert '!d' in second_message
    assert len(testbot.bot.flow_executor.in_flight) == 1
    assert testbot.bot.flow_executor.in_flight[0].name == 'w2'


def test_manual_flow(testbot):
    assert 'Flow w1 started' in testbot.exec_command('!flows start w1')
    flow_message = testbot.pop_message()
    assert 'You are in the flow w1, you can continue with' in flow_message
    assert '!a' in flow_message
    assert 'a' in testbot.exec_command('!a')
    flow_message = testbot.pop_message()
    assert 'You are in the flow w1, you can continue with' in flow_message
    assert '!b' in flow_message
    assert '!c' in flow_message


def test_manual_flow_with_or_without_hinting(testbot):
    assert 'Flow w4 started' in testbot.exec_command('!flows start w4')
    assert 'a' in testbot.exec_command('!a')
    assert 'b' in testbot.exec_command('!b')
    flow_message = testbot.pop_message()
    assert 'You are in the flow w4, you can continue with' in flow_message
    assert '!c' in flow_message


def test_no_flyby_trigger_flow(testbot):
    testbot.bot.push_message('!flows start w1')
    # One message or the other can arrive first.
    flow_message = testbot.pop_message()
    assert 'Flow w1 started' in flow_message or 'You are in the flow w1' in flow_message
    flow_message = testbot.pop_message()
    assert 'Flow w1 started' in flow_message or 'You are in the flow w1' in flow_message
    assert 'a' in testbot.exec_command('!a')
    flow_message = testbot.pop_message()
    assert 'You are in the flow w1' in flow_message

    assert 'c' in testbot.exec_command('!c')  # c is a trigger for w2 but it should not trigger now.
    flow_message = testbot.pop_message()
    assert 'You are in the flow w1' in flow_message
    assert len(testbot.bot.flow_executor.in_flight) == 1


def test_flow_only(testbot):
    assert 'a' in testbot.exec_command('!a')  # non flow_only should respond.
    testbot.push_message('!d')
    with pytest.raises(Empty):
        testbot.pop_message(timeout=1)


def test_flow_only_help(testbot):
    testbot.push_message('!help')
    msg = testbot.bot.pop_message()
    assert '!a' in msg  # non flow_only should be in help by default
    assert '!d' not in msg  # flow_only should not be in help by default


def test_flows_stop(testbot):
    assert 'c' in testbot.exec_command('!c')
    flow_message = testbot.bot.pop_message()
    assert 'You are in the flow w2' in flow_message
    assert 'w2 stopped' in testbot.exec_command('!flows stop w2')
    assert len(testbot.bot.flow_executor.in_flight) == 0


def test_flows_kill(testbot):
    assert 'c' in testbot.exec_command('!c')
    flow_message = testbot.bot.pop_message()
    assert 'You are in the flow w2' in flow_message
    assert 'w2 killed' in testbot.exec_command('!flows kill gbin@localhost w2')


def test_room_flow(testbot):
    assert 'Flow w3 started' in testbot.exec_command('!flows start w3')
    flow_message = testbot.pop_message()
    assert 'You are in the flow w3, you can continue with' in flow_message
    assert '!a' in flow_message
    assert 'a' in testbot.exec_command('!a')
    flow_message = testbot.pop_message()
    assert 'You are in the flow w3, you can continue with' in flow_message
    assert '!b' in flow_message



import inspect
import logging
import sys

COLORS = {'DEBUG': 'cyan', 'INFO': 'green', 'WARNING': 'yellow', 'ERROR': 'red', 'CRITICAL': 'red', }

NO_COLORS = {'DEBUG': '', 'INFO': '', 'WARNING': '', 'ERROR': '', 'CRITICAL': '', }


def ispydevd():
    for frame in inspect.stack():
        if frame[1].endswith("pydevd.py"):
            return True
    return False


root_logger = logging.getLogger()
root_logger.setLevel(logging.INFO)

pydev = ispydevd()
stream = sys.stdout if pydev else sys.stderr
isatty = pydev or stream.isatty()  # force isatty if we are under pydev because it supports coloring anyway.
console_hdlr = logging.StreamHandler(stream)


def get_log_colors(theme_color=None):
    """Return a tuple containing the log format string and a log color dict"""
    if theme_color == 'light':
        text_color_theme = 'white'
    elif theme_color == 'dark':
        text_color_theme = 'black'
    else:  # Anything else produces nocolor
        return '%(name)-25.25s%(reset)s %(message)s%(reset)s', NO_COLORS

    return f'%(name)-25.25s%(reset)s %({text_color_theme})s%(message)s%(reset)s', COLORS


def format_logs(formatter=None, theme_color=None):
    """
    You may either use the formatter parameter to provide your own
    custom formatter, or the theme_color parameter to use the
    built in color scheme formatter.
    """
    if formatter:
        console_hdlr.setFormatter(formatter)
    # if isatty and not True:
    elif isatty:
        from colorlog import ColoredFormatter  # noqa
        log_format, colors_dict = get_log_colors(theme_color)
        color_formatter = ColoredFormatter(
            "%(asctime)s %(log_color)s%(levelname)-8s%(reset)s " + log_format,
            datefmt="%H:%M:%S",
            reset=True,
            log_colors=colors_dict,
        )
        console_hdlr.setFormatter(color_formatter)
    else:
        console_hdlr.setFormatter(logging.Formatter("%(asctime)s %(levelname)-8s %(name)-25s %(message)s"))
    root_logger.addHandler(console_hdlr)



import logging
import re
import sys
from functools import lru_cache

from errbot.backends.base import Room, RoomDoesNotExistError, RoomOccupant, Stream
from errbot.backends.xmpp import XMPPRoomOccupant, XMPPBackend, XMPPConnection, split_identifier

from markdown import Markdown
from markdown.extensions.extra import ExtraExtension
from markdown.extensions import Extension
from markdown.treeprocessors import Treeprocessor

from email.mime.multipart import MIMEMultipart
import email.mime.application

import requests

log = logging.getLogger(__name__)

try:
    import hypchat
except ImportError:
    log.exception("Could not start the HipChat backend")
    log.fatal(
        "You need to install the hipchat support in order to use the HipChat backend.\n "
        "You should be able to install this package using:\n"
        "pip install errbot[hipchat]"
    )
    sys.exit(1)

COLORS = {
    'blue': 'purple',
    'white': 'gray',
    'black': 'gray',
}  # best effort to map errbot colors to hipchat ones,


# Rendering customizations
class HipchatTreeprocessor(Treeprocessor):
    def run(self, root):
        def recurse_patch(element):
            t = element.tag
            if t == 'h1':
                element.tag = 'strong'
                element.text = element.text.upper()
            elif t == 'h2':
                element.tag = 'em'
            elif t in ('h3', 'h4', 'h5', 'h6'):
                element.tag = 'p'
            elif t == 'hr':
                element.tag = 'p'
                element.text = '─' * 80

            for elems in element:
                recurse_patch(elems)
        recurse_patch(root)


class HipchatExtension(Extension):
    """Removes the unsupported html tags from hipchat"""

    def extendMarkdown(self, md, md_globals):
        md.registerExtension(self)
        md.treeprocessors.add("hipchat stripper", HipchatTreeprocessor(), '<inline')
        log.debug("Will apply those treeprocessors:\n%s", md.treeprocessors)


def hipchat_html():
    return Markdown(output_format='xhtml', extensions=[ExtraExtension(), HipchatExtension()])


class HipChatRoomOccupant(XMPPRoomOccupant):
    """
    An occupant of a Multi-User Chatroom.

    This class has all the attributes that are returned by a call to
    https://www.hipchat.com/docs/apiv2/method/get_all_participants
    with the link to self expanded.
    """
    def __init__(self, node=None, domain=None, resource=None, room=None, hipchat_user=None, aclattr=None):
        """
        :param hipchat_user:
            A user object as returned by
            https://www.hipchat.com/docs/apiv2/method/get_all_participants
            with the link to self expanded.
        """
        if hipchat_user:
            for k, v in hipchat_user.items():
                setattr(self, k, v)
            # Quick fix to be able to all the parent.
            if '/' in hipchat_user['xmpp_jid']:
                node_domain, resource = hipchat_user['xmpp_jid'].split('/')
            else:
                node_domain = hipchat_user['xmpp_jid']
                resource = hipchat_user['name']
            node, domain = node_domain.split('@')

        self._aclattr = aclattr

        super().__init__(node, domain, resource, room)

    @property
    def aclattr(self):
        return self._aclattr


class HipChatRoom(Room):
    """
    This class represents a Multi-User Chatroom.
    """

    def __init__(self, name, bot):
        """
        :param name:
            The name of the room
        """
        self.name = name
        self.hypchat = bot.conn.hypchat
        self.xep0045 = bot.conn.client.plugin['xep_0045']
        self._bot = bot

    @property
    def room(self):
        """
        Return room information from the HipChat API
        """
        try:
            log.debug('Querying HipChat API for room %s.', self.name)
            return self.hypchat.get_room(self.name)
        except hypchat.requests.HttpNotFound:
            raise RoomDoesNotExistError('The given room does not exist.')

    @property
    def name(self):
        """
        The name of this room
        """
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def jid(self):
        return self.room['xmpp_jid']

    @property
    def node(self):
        return self._bot.build_identifier(self.jid).node

    @property
    def domain(self):
        return self._bot.build_identifier(self.jid).domain

    @property
    def resource(self):
        return self._bot.build_identifier(self.jid).resource

    def __repr__(self):
        return f"<HipChatMUCRoom('{self.name}')>"

    def __str__(self):
        return self.room['xmpp_jid']

    def join(self, username=None, password=None):
        """
        Join the room.

        If the room does not exist yet, this will automatically call
        :meth:`create` on it first.
        """
        if not self.exists:
            self.create()

        room = self.jid
        self.xep0045.joinMUC(room, username, password=password, wait=True)
        self._bot.conn.add_event_handler(f'muc::{room}::got_online', self._bot.user_joined_chat)
        self._bot.conn.add_event_handler(f'muc::{room}::got_offline', self._bot.user_left_chat)

        self._bot.callback_room_joined(self)
        log.info('Joined room %s.', self.name)

    def leave(self, reason=None):
        """
        Leave the room.

        :param reason:
            An optional string explaining the reason for leaving the room
        """
        if reason is None:
            reason = ""
        room = self.jid
        try:
            self.xep0045.leaveMUC(room=room, nick=self.xep0045.ourNicks[room], msg=reason)
            self._bot.conn.del_event_handler(f'muc::{room}::got_online', self._bot.user_joined_chat)
            self._bot.conn.del_event_handler(f'muc::{room}::got_offline', self._bot.user_left_chat)
            log.info('Left room %s', self)
            self._bot.callback_room_left(self)
        except KeyError:
            log.debug('Trying to leave %s while not in this room.', self)

    def create(self, privacy='public', guest_access=False):
        """
        Create the room.

        Calling this on an already existing room is a no-op.

        :param privacy:
            Whether the room is available for access by other users or not.
            Valid values are "public" and "private".
        :param guest_access:
            Whether or not to enable guest access for this room.
        """
        if self.exists:
            log.debug('Tried to create the room %s, but it has already been created', self)
        else:
            self.hypchat.create_room(
                name=self.name,
                privacy=privacy,
                guest_access=guest_access
            )
            log.info('Created room %s.', self)

    def destroy(self):
        """
        Destroy the room.

        Calling this on a non-existing room is a no-op.
        """
        try:
            self.room.delete()
            log.info('Destroyed room %s.', self)
        except RoomDoesNotExistError:
            log.debug("Can't destroy room %s, it doesn't exist.", self)

    @property
    def exists(self):
        """
        Boolean indicating whether this room already exists or not.

        :getter:
            Returns `True` if the room exists, `False` otherwise.
        """
        try:
            self.hypchat.get_room(self.name)
            return True
        except hypchat.requests.HttpNotFound:
            return False

    @property
    def joined(self):
        """
        Boolean indicating whether this room has already been joined or not.

        :getter:
            Returns `True` if the room has been joined, `False` otherwise.
        """
        return self.jid in self.xep0045.getJoinedRooms()

    @property
    def topic(self):
        """
        The room topic.

        :getter:
            Returns the topic (a string) if one is set, `None` if no
            topic has been set at all.
        """
        return self.room['topic']

    @topic.setter
    def topic(self, topic):
        """
        Set the room's topic.

        :param topic:
            The topic to set.
        """
        self.room.topic(topic)
        log.debug('Changed topic of %s to %s', self, topic)

    @property
    def occupants(self):
        """
        The room's occupants.

        :getter:
            Returns a list of :class:`~HipChatMUCOccupant` instances.
        """
        participants = self.room.participants(expand='items')['items']
        occupants = []
        for p in participants:
            occupants.append(HipChatRoomOccupant(hipchat_user=p))
        return occupants

    def invite(self, *args):
        """
        Invite one or more people into the room.

        :param args:
            One or more people to invite into the room. May be the
            mention name (beginning with an @) or "FirstName LastName"
            of the user you wish to invite.
        """
        room = self.room
        users = self._bot.conn.users

        for person in args:
            try:
                if person.startswith('@'):
                    user = [u for u in users if u['mention_name'] == person[1:]][0]
                else:
                    user = [u for u in users if u['name'] == person][0]
            except IndexError:
                logging.warning('No user by the name of %s found.', person)
            else:
                if room['privacy'] == 'private':
                    room.members().add(user)
                    log.info('Added %s to private room %s.', user['name'], self)
                room.invite(user, 'No reason given.')
                log.info('Invited %s to %s.', person, self)

    def notify(self, message, color=None, notify=False, message_format=None):
        """
        Send a notification to a room.

        See the
        `HipChat API documentation <https://www.hipchat.com/docs/apiv2/method/send_room_notification>`_
        for more info.
        """
        self.room.notification(message=message, color=color, notify=notify, format=message_format)


class HipchatClient(XMPPConnection):
    def __init__(self, *args, **kwargs):
        self.token = kwargs.pop('token')
        self.endpoint = kwargs.pop('endpoint')
        self._cached_users = None
        verify = kwargs.pop('verify')
        if verify is None:
            verify = True
        if self.endpoint is None:
            self.hypchat = hypchat.HypChat(self.token, verify=verify)
        else:
            # We could always pass in the endpoint, with a default value if it's
            # None, but this way we support hypchat<0.18
            self.hypchat = hypchat.HypChat(self.token, endpoint=self.endpoint, verify=verify)
        super().__init__(*args, **kwargs)

    @property
    def users(self):
        """
        A list of all the users.

        See also: https://www.hipchat.com/docs/apiv2/method/get_all_users
        """

        if not self._cached_users:
            result = self.hypchat.users(guests=True)
            users = result['items']
            next_link = 'next' in result['links']
            while next_link:
                result = result.next()
                users += result['items']
                next_link = 'next' in result['links']
            self._cached_users = users
        return self._cached_users


class HipchatBackend(XMPPBackend):
    room_factory = HipChatRoom
    roomoccupant_factory = HipChatRoomOccupant

    def __init__(self, config):
        self.api_token = config.BOT_IDENTITY['token']
        self.api_endpoint = config.BOT_IDENTITY.get('endpoint', None)
        self.api_verify = config.BOT_IDENTITY.get('verify', True)
        self.md = hipchat_html()
        super().__init__(config)

    def create_connection(self):
        # HipChat connections time out with the default keepalive interval
        # so use a lower value that is known to work, but only if the user
        # does not specify their own value in their config.
        if self.keepalive is None:
            self.keepalive = 60

        return HipchatClient(
            jid=self.jid,
            password=self.password,
            feature=self.feature,
            keepalive=self.keepalive,
            ca_cert=self.ca_cert,
            token=self.api_token,
            endpoint=self.api_endpoint,
            server=self.server,
            verify=self.api_verify,
        )

    def _build_room_occupant(self, txtrep):
        node, domain, resource = split_identifier(txtrep)
        return self.roomoccupant_factory(node,
                                         domain,
                                         resource,
                                         self.query_room(node + '@' + domain),
                                         aclattr=self._find_user(resource, 'name'))

    def callback_message(self, msg):
        super().callback_message(msg)
        possible_mentions = re.findall(r'@\w+', msg.body)
        people = list(
            filter(None.__ne__, [self._find_user(mention[1:], 'mention_name') for mention in possible_mentions])
        )

        if people:
            self.callback_mention(msg, people)

    @property
    def mode(self):
        return 'hipchat'

    def rooms(self):
        """
        Return a list of rooms the bot is currently in.

        :returns:
            A list of :class:`~HipChatRoom` instances.
        """
        xep0045 = self.conn.client.plugin['xep_0045']
        rooms = {}
        # Build a mapping of xmpp_jid->name for easy reference
        for room in self.conn.hypchat.rooms(expand='items').contents():
            rooms[room['xmpp_jid']] = room['name']

        joined_rooms = []
        for room in xep0045.getJoinedRooms():
            try:
                joined_rooms.append(HipChatRoom(rooms[room], self))
            except KeyError:
                pass
        return joined_rooms

    @lru_cache(1024)
    def query_room(self, room):
        """
        Query a room for information.

        :param room:
            The name (preferred) or XMPP JID of the room to query for.
        :returns:
            An instance of :class:`~HipChatRoom`.
        """
        if room.endswith('@conf.hipchat.com') or room.endswith('@conf.btf.hipchat.com'):
            log.debug("Room specified by JID, looking up room name")
            rooms = self.conn.hypchat.rooms(expand='items').contents()
            try:
                name = [r['name'] for r in rooms if r['xmpp_jid'] == room][0]
            except IndexError:
                raise RoomDoesNotExistError(f'No room with JID {room} found.')
            log.info('Found %s to be the room %s, consider specifying this directly.', room, name)
        else:
            name = room

        return HipChatRoom(name, self)

    def build_reply(self, msg, text=None, private=False, threaded=False):
        response = super().build_reply(msg=msg, text=text, private=private, threaded=threaded)
        if msg.is_group and msg.frm == response.to:
            # HipChat violates the XMPP spec :( This results in a valid XMPP JID
            # but HipChat mangles them into stuff like
            # "132302_961351@chat.hipchat.com/none||proxy|pubproxy-b100.hipchat.com|5292"
            # so we request the user's proper JID through their API and use that here
            # so that private responses originating from a room (IE, DIVERT_TO_PRIVATE)
            # work correctly.
            response.to = self._find_user(response.to.client, 'name')
        return response

    def send_card(self, card):
        if isinstance(card.to, RoomOccupant):
            card.to = card.to.room
        if not card.is_group:
            raise ValueError('Private notifications/cards are impossible to send on 1 to 1 messages on hipchat.')
        log.debug('room id = %s', card.to)
        room = self.query_room(str(card.to)).room

        data = {'message': '-' if not card.body else self.md.convert(card.body),
                'notify': False,
                'message_format': 'html'}

        if card.color:
            data['color'] = COLORS[card.color] if card.color in COLORS else card.color

        hcard = {'id': f'FF{card.__hash__():0.16X}'}

        # Only title is supported all across the types.
        if card.title:
            hcard['title'] = card.title
        else:
            hcard['title'] = ' '  # title is mandatory, more that 1 chr.

        # Go from the most restrictive type to the less resctrictive to find the most appropriate.
        if card.image and not card.summary and not card.fields and not card.link:
            hcard['style'] = 'image'
            hcard['thumbnail'] = {'url': card.image if not card.thumbnail else card.thumbnail}
            hcard['url'] = card.image
            if card.body:
                data['message'] = card.body  # We don't have a card body field so retrofit it to the main body.
        elif card.link and not card.summary and not card.fields:
            hcard['style'] = 'link'
            hcard['url'] = card.link
            if card.thumbnail:
                hcard['icon'] = {'url': card.thumbnail}
            if card.image:
                hcard['thumbnail'] = {'url': card.image}
            if card.body:
                hcard['description'] = card.body
        else:
            hcard['style'] = 'application'
            hcard['format'] = 'medium'
            if card.image and card.thumbnail:
                log.warning('Hipchat cannot display this card with an image.'
                            'Remove summary, fields and/or possibly link to fallback to an hichat link or '
                            'an image style card.')
            if card.image or card.thumbnail:
                hcard['icon'] = {'url': card.thumbnail if card.thumbnail else card.image}
            if card.body:
                hcard['description'] = card.body
            if card.summary:
                hcard['activity'] = {'html': card.summary}
            if card.fields:
                hcard['attributes'] = [{'label': key, 'value': {'label': value, 'style': 'lozenge-complete'}}
                                       for key, value in card.fields]
            if card.link:
                hcard['url'] = card.link

        data['card'] = hcard

        log.debug("Sending request:" + str(data))
        room._requests.post(room.url + '/notification', data=data)  # noqa

    def send_stream_request(self, identifier, fsource, name='file.txt', size=None, stream_type=None):
        """Starts a file transfer.
            note, fsource used to make the stream needs to be in open/rb state
        """

        stream = Stream(identifier=identifier, fsource=fsource, name=name, size=size, stream_type=stream_type)
        result = self.thread_pool.apply_async(self._hipchat_upload, (stream,))
        log.debug('Response from server: %s', result.get(timeout=10))
        return stream

    def _hipchat_upload(self, stream):
        """ Uploads file in a stream  """
        try:
            stream.accept()
            room = self.query_room(str(stream.identifier)).room
            headers = {
                'Authorization': f'Bearer {self.api_token}',
                'Accept-Charset': 'UTF-8',
                'Content-Type': 'multipart/related',
            }
            raw_body = MIMEMultipart('related')
            img = email.mime.application.MIMEApplication(stream.read())
            img.add_header(
                'Content-Disposition',
                'attachment',
                name='file',
                filename=stream.name
            )
            raw_body.attach(img)
            raw_headers, body = raw_body.as_string().split('\n\n', 1)
            boundary = re.search('boundary="([^\"]*)"', raw_headers).group(1)
            headers['Content-Type'] = f'multipart/related; boundary="{boundary}"'
            resp = requests.post(room.url + '/share/file', headers=headers, data=body)
            log.info('Request ok: %s.', resp.ok)

            if resp.ok:
                log.info('Request status: %s.', resp.status_code)
                stream.success()
            else:
                log.error('Request status: %s.', resp.status_code)
                log.error('Request reason: %s.', resp.reason)
                log.error('Request text: %s.', resp.text)
                stream.error()
        except Exception:
            log.exception(f'Upload of {stream.name} to {stream.identifier.channelname} failed.')

    @lru_cache(1024)
    def _find_user(self, name, criteria):
        """
        Find a specific hipchat user with a simple criteria like 'name' or 'mention_name' and returns
        its jid.

        :param name: the value you seek.
        :param criteria: 'name' or 'mention_name'
        :return: the matching XMPPPerson or None if not found.
        """
        users = [u for u in self.conn.users if u[criteria] == name]
        if not users:
            log.debug('Failed to find user %s', name)
            return None
        userdetail = self.conn.hypchat.get_user(users[0]['id'])
        identifier = self.build_identifier(userdetail['xmpp_jid'])

        return identifier

    def prefix_groupchat_reply(self, message, identifier):
        message.body = f'@{identifier.nick}: {message.body}'

    def __hash__(self):
        return 0  # it is a singleton anyway



from errbot import BotPlugin


class MentionTestPlugin(BotPlugin):
    def callback_mention(self, msg, people):
        if self.bot_identifier in people:
            self.send(msg.frm, "Somebody mentioned me!", msg)
            return
        self.send(msg.frm, "Somebody mentioned %s!" % ','.join(p.person for p in people), msg)



# vim: noai:ts=4:sw=4
import re

from markdown import Markdown
from markdown.extensions.extra import ExtraExtension

# Attribute regexp looks for extendend syntax: {: ... }
ATTR_RE = re.compile(r'{:([^}]*)}')
MD_ESCAPE_RE = re.compile('|'.join(re.escape(c) for c in ('\\', '`', '*', '_', '{', '}', '[', ']',
                                                          '(', ')', '>', '#', '+', '-', '.', '!')))

# Here are few helpers to simplify the conversion from markdown to various
# backend formats.


def ansi():
    """This makes a converter from markdown to ansi (console) format.
    It can be called like this:
    from errbot.rendering import ansi
    md_converter = ansi()  # you need to cache the converter

    ansi_txt = md_converter.convert(md_txt)
    """
    from .ansiext import AnsiExtension
    md = Markdown(output_format='ansi', extensions=[ExtraExtension(), AnsiExtension()])
    md.stripTopLevelTags = False
    return md


def text():
    """This makes a converter from markdown to text (unicode) format.
    It can be called like this:
    from errbot.rendering import text
    md_converter = text()  # you need to cache the converter

    pure_text = md_converter.convert(md_txt)
    """
    from .ansiext import AnsiExtension
    md = Markdown(output_format='text', extensions=[ExtraExtension(), AnsiExtension()])
    md.stripTopLevelTags = False
    return md


def imtext():
    """This makes a converter from markdown to imtext (unicode) format.
    imtest is the format like gtalk, slack or skype with simple _ or * markup.

    It can be called like this:
    from errbot.rendering import imtext
    md_converter = imtext()  # you need to cache the converter

    im_text = md_converter.convert(md_txt)
    """
    from .ansiext import AnsiExtension
    md = Markdown(output_format='imtext', extensions=[ExtraExtension(), AnsiExtension()])
    md.stripTopLevelTags = False
    return md


class Mde2mdConverter(object):
    def convert(self, mde):
        while True:
            m = ATTR_RE.search(mde)
            if m is None:
                break
            left, right = m.span()
            mde = mde[:left] + mde[right:]
        return mde


def md():
    """This makes a converter from markdown-extra to markdown, stripping the attributes from extra.
    """
    return Mde2mdConverter()


def xhtml():
    """This makes a converter from markdown to xhtml format.
    It can be called like this:
    from errbot.rendering import xhtml
    md_converter = xhtml()  # you need to cache the converter

    html = md_converter.convert(md_txt)
    """
    return Markdown(output_format='xhtml', extensions=[ExtraExtension()])


def md_escape(txt):
    """ Call this if you want to be sure your text won't be interpreted as markdown
    :param txt: bare text to escape.
    """
    return MD_ESCAPE_RE.sub(lambda match: '\\' + match.group(0), txt)



from inspect import getmembers, ismethod
from json import loads
import logging

from flask.app import Flask
from flask.views import View
from flask import request
import errbot.core_plugins

log = logging.getLogger(__name__)


def strip_path():
    # strip the trailing slashes on incoming requests
    request.environ['PATH_INFO'] = request.environ['PATH_INFO'].rstrip('/')


def try_decode_json(req):
    data = req.data.decode()
    try:
        return loads(data)
    except Exception:
        return None


def reset_app():
    """Zap everything here, useful for unit tests
    """
    errbot.core_plugins.flask_app = Flask(__name__)


def route(obj):
    """Check for functions to route in obj and route them."""
    flask_app = errbot.core_plugins.flask_app
    classname = obj.__class__.__name__
    log.info("Checking %s for webhooks", classname)
    for name, func in getmembers(obj, ismethod):
        if getattr(func, '_err_webhook_uri_rule', False):
            log.info("Webhook routing %s", func.__name__)
            form_param = func._err_webhook_form_param
            uri_rule = func._err_webhook_uri_rule
            verbs = func._err_webhook_methods
            raw = func._err_webhook_raw

            callable_view = WebView.as_view(func.__name__ + '_' + '_'.join(verbs), func, form_param, raw)

            # Change existing rule.
            for rule in flask_app.url_map._rules:
                if rule.rule == uri_rule:
                    flask_app.view_functions[rule.endpoint] = callable_view
                    return

            # Add a new rule
            flask_app.add_url_rule(uri_rule, view_func=callable_view, methods=verbs, strict_slashes=False)


class WebView(View):
    def __init__(self, func, form_param, raw):
        if form_param is not None and raw:
            raise Exception("Incompatible parameters: form_param cannot be set if raw is True")
        self.func = func
        self.raw = raw
        self.form_param = form_param
        self.method_filter = lambda obj: ismethod(obj) and self.func.__name__ == obj.__name__

    def dispatch_request(self, *args, **kwargs):

        if self.raw:  # override and gives the request directly
            response = self.func(request, **kwargs)
        elif self.form_param:
            content = request.form.get(self.form_param)
            if content is None:
                raise Exception('Received a request on a webhook with a form_param defined, '
                                'but that key (%s) is missing from the request.', self.form_param)
            try:
                content = loads(content)
            except ValueError:
                log.debug('The form parameter is not JSON, return it as a string.')
            response = self.func(content, **kwargs)
        else:
            data = try_decode_json(request)
            if not data:
                if hasattr(request, 'forms'):
                    data = dict(request.forms)  # form encoded
                else:
                    data = request.data.decode()
            response = self.func(data, **kwargs)
        return response if response else ''  # assume None as an OK response (simplifies the client side)



import logging
from time import sleep
from errbot.backends.base import ONLINE

from errbot.backends.test import TestPerson
from errbot.core import ErrBot

log = logging.getLogger(__name__)


class ConnectionMock(object):
    def send(self, msg):
        pass

    def send_message(self, msg):
        pass


class NullBackend(ErrBot):
    conn = ConnectionMock()
    running = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.bot_identifier = self.build_identifier('Err')  # whatever

    def serve_forever(self):
        self.connect()  # be sure we are "connected" before the first command
        self.connect_callback()  # notify that the connection occured
        try:
            while self.running:
                sleep(1)

        except EOFError:
            pass
        except KeyboardInterrupt:
            pass
        finally:
            log.debug("Trigger disconnect callback")
            self.disconnect_callback()
            log.debug("Trigger shutdown")
            self.shutdown()

    def connect(self):
        if not self.conn:
            self.conn = ConnectionMock()
        return self.conn

    def build_identifier(self, strrep):
        return TestPerson(strrep)

    def shutdown(self):
        if self.running:
            self.running = False
            super().shutdown()  # only once (hackish)

    def change_presence(self, status: str = ONLINE, message: str = '') -> None:
        pass

    def build_reply(self, msg, text=None, private=False, threaded=False):
        pass

    def prefix_groupchat_reply(self, message, identifier):
        pass

    def query_room(self, room):
        pass

    def rooms(self):
        pass

    @property
    def mode(self):
        return 'null'
